# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: AccountSummary.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import APIRequest_pb2 as APIRequest__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14\x41\x63\x63ountSummary.proto\x12\x0e\x41\x63\x63ountSummary\x1a\x10\x41PIRequest.proto\"N\n\x15\x41\x63\x63ountSummaryRequest\x12\x16\n\x0esummaryVersion\x18\x01 \x01(\x05\x12\x1d\n\x15includeRecentActivity\x18\x02 \x01(\x08\"\x98\x05\n\x16\x41\x63\x63ountSummaryElements\x12\x11\n\tclientKey\x18\x01 \x01(\x0c\x12*\n\x08settings\x18\x02 \x01(\x0b\x32\x18.AccountSummary.Settings\x12*\n\x08keysInfo\x18\x03 \x01(\x0b\x32\x18.AccountSummary.KeysInfo\x12)\n\x08syncLogs\x18\x04 \x03(\x0b\x32\x17.AccountSummary.SyncLog\x12\x19\n\x11isEnterpriseAdmin\x18\x05 \x01(\x08\x12(\n\x07license\x18\x06 \x01(\x0b\x32\x17.AccountSummary.License\x12$\n\x05group\x18\x07 \x01(\x0b\x32\x15.AccountSummary.Group\x12\x32\n\x0c\x45nforcements\x18\x08 \x01(\x0b\x32\x1c.AccountSummary.Enforcements\x12(\n\x06Images\x18\t \x03(\x0b\x32\x18.AccountSummary.KeyValue\x12\x30\n\x0fpersonalLicense\x18\n \x01(\x0b\x32\x17.AccountSummary.License\x12\x1e\n\x16\x66ixSharedFolderRecords\x18\x0b \x01(\x08\x12\x11\n\tusernames\x18\x0c \x03(\t\x12+\n\x07\x64\x65vices\x18\r \x03(\x0b\x32\x1a.AccountSummary.DeviceInfo\x12\x14\n\x0cisShareAdmin\x18\x0e \x01(\x08\x12\x17\n\x0f\x61\x63\x63ountRecovery\x18\x0f \x01(\x08\x12\x1d\n\x15\x61\x63\x63ountRecoveryPrompt\x18\x10 \x01(\x08\x12\'\n\x1fminMasterPasswordLengthNoPrompt\x18\x11 \x01(\x05\x12\x16\n\x0e\x66orbidKeyType2\x18\x12 \x01(\x08\"\x8b\x03\n\nDeviceInfo\x12\x1c\n\x14\x65ncryptedDeviceToken\x18\x01 \x01(\x0c\x12\x12\n\ndeviceName\x18\x02 \x01(\t\x12\x32\n\x0c\x64\x65viceStatus\x18\x03 \x01(\x0e\x32\x1c.Authentication.DeviceStatus\x12\x17\n\x0f\x64\x65vicePublicKey\x18\x04 \x01(\x0c\x12 \n\x18\x65ncryptedDataKeyDoNotUse\x18\x05 \x01(\x0c\x12\x15\n\rclientVersion\x18\x06 \x01(\t\x12\x10\n\x08username\x18\x07 \x01(\t\x12\x11\n\tipAddress\x18\x08 \x01(\t\x12\x1a\n\x12\x61pproveRequestTime\x18\t \x01(\x03\x12\x1f\n\x17\x65ncryptedDataKeyPresent\x18\n \x01(\x08\x12\x0f\n\x07groupId\x18\x0b \x01(\x03\x12\x16\n\x0e\x64\x65vicePlatform\x18\x0c \x01(\t\x12:\n\x10\x63lientFormFactor\x18\r \x01(\x0e\x32 .Authentication.ClientFormFactor\"\xc1\x01\n\x08KeysInfo\x12\x18\n\x10\x65ncryptionParams\x18\x01 \x01(\x0c\x12\x18\n\x10\x65ncryptedDataKey\x18\x02 \x01(\x0c\x12\x19\n\x11\x64\x61taKeyBackupDate\x18\x03 \x01(\x01\x12\x13\n\x0buserAuthUid\x18\x04 \x01(\x0c\x12\x1b\n\x13\x65ncryptedPrivateKey\x18\x05 \x01(\x0c\x12\x1e\n\x16\x65ncryptedEccPrivateKey\x18\x06 \x01(\x0c\x12\x14\n\x0c\x65\x63\x63PublicKey\x18\x07 \x01(\x0c\"\x81\x01\n\x07SyncLog\x12\x13\n\x0b\x63ountryName\x18\x01 \x01(\t\x12\x12\n\nsecondsAgo\x18\x02 \x01(\x03\x12\x12\n\ndeviceName\x18\x03 \x01(\t\x12\x13\n\x0b\x63ountryCode\x18\x04 \x01(\t\x12\x11\n\tdeviceUID\x18\x05 \x01(\x0c\x12\x11\n\tipAddress\x18\x06 \x01(\t\"\xe7\x06\n\x07License\x12\x18\n\x10subscriptionCode\x18\x01 \x01(\t\x12\x15\n\rproductTypeId\x18\x02 \x01(\x05\x12\x17\n\x0fproductTypeName\x18\x03 \x01(\t\x12\x16\n\x0e\x65xpirationDate\x18\x04 \x01(\t\x12\x1e\n\x16secondsUntilExpiration\x18\x05 \x01(\x03\x12\x12\n\nmaxDevices\x18\x06 \x01(\x05\x12\x14\n\x0c\x66ilePlanType\x18\x07 \x01(\x05\x12\x11\n\tbytesUsed\x18\x08 \x01(\x03\x12\x12\n\nbytesTotal\x18\t \x01(\x03\x12%\n\x1dsecondsUntilStorageExpiration\x18\n \x01(\x03\x12\x1d\n\x15storageExpirationDate\x18\x0b \x01(\t\x12,\n$hasAutoRenewableAppstoreSubscription\x18\x0c \x01(\x08\x12\x13\n\x0b\x61\x63\x63ountType\x18\r \x01(\x05\x12\x18\n\x10uploadsRemaining\x18\x0e \x01(\x05\x12\x14\n\x0c\x65nterpriseId\x18\x0f \x01(\x05\x12\x13\n\x0b\x63hatEnabled\x18\x10 \x01(\x08\x12 \n\x18\x61uditAndReportingEnabled\x18\x11 \x01(\x08\x12!\n\x19\x62reachWatchFeatureDisable\x18\x12 \x01(\x08\x12\x12\n\naccountUid\x18\x13 \x01(\x0c\x12\x1c\n\x14\x61llowPersonalLicense\x18\x14 \x01(\x08\x12\x12\n\nlicensedBy\x18\x15 \x01(\t\x12\r\n\x05\x65mail\x18\x16 \x01(\t\x12\x1a\n\x12\x62reachWatchEnabled\x18\x17 \x01(\x08\x12\x1a\n\x12\x62reachWatchScanned\x18\x18 \x01(\x08\x12\x1d\n\x15\x62reachWatchExpiration\x18\x19 \x01(\x03\x12\x1e\n\x16\x62reachWatchDateCreated\x18\x1a \x01(\x03\x12%\n\x05\x65rror\x18\x1b \x01(\x0b\x32\x16.AccountSummary.Result\x12\x12\n\nexpiration\x18\x1d \x01(\x03\x12\x19\n\x11storageExpiration\x18\x1e \x01(\x03\x12\x14\n\x0cuploadsCount\x18\x1f \x01(\x05\x12\r\n\x05units\x18  \x01(\x05\x12\x19\n\x11pendingEnterprise\x18! \x01(\x08\x12\x14\n\x0cisPamEnabled\x18\" \x01(\x08\"\xa3\x01\n\x05\x41\x64\x64On\x12\x14\n\x0clicenseKeyId\x18\x01 \x01(\x05\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x16\n\x0e\x65xpirationDate\x18\x03 \x01(\x03\x12\x13\n\x0b\x63reatedDate\x18\x04 \x01(\x03\x12\x0f\n\x07isTrial\x18\x05 \x01(\x08\x12\x0f\n\x07\x65nabled\x18\x06 \x01(\x08\x12\x0f\n\x07scanned\x18\x07 \x01(\x08\x12\x16\n\x0e\x66\x65\x61tureDisable\x18\x08 \x01(\x08\"\xbd\t\n\x08Settings\x12\r\n\x05\x61udit\x18\x01 \x01(\x08\x12!\n\x19mustPerformAccountShareBy\x18\x02 \x01(\x03\x12>\n\x0eshareAccountTo\x18\x03 \x03(\x0b\x32&.AccountSummary.MissingAccountShareKey\x12+\n\x05rules\x18\x04 \x03(\x0b\x32\x1c.AccountSummary.PasswordRule\x12\x1a\n\x12passwordRulesIntro\x18\x05 \x01(\t\x12\x16\n\x0e\x61utoBackupDays\x18\x06 \x01(\x05\x12\r\n\x05theme\x18\x07 \x01(\t\x12\x0f\n\x07\x63hannel\x18\x08 \x01(\t\x12\x14\n\x0c\x63hannelValue\x18\t \x01(\t\x12\x15\n\rrsaConfigured\x18\n \x01(\x08\x12\x15\n\remailVerified\x18\x0b \x01(\x08\x12\"\n\x1amasterPasswordLastModified\x18\x0c \x01(\x01\x12\x18\n\x10\x61\x63\x63ountFolderKey\x18\r \x01(\x0c\x12\x31\n\x0csecurityKeys\x18\x0e \x03(\x0b\x32\x1b.AccountSummary.SecurityKey\x12+\n\tkeyValues\x18\x0f \x03(\x0b\x32\x18.AccountSummary.KeyValue\x12\x0f\n\x07ssoUser\x18\x10 \x01(\x08\x12\x18\n\x10onlineAccessOnly\x18\x11 \x01(\x08\x12\x1c\n\x14masterPasswordExpiry\x18\x12 \x01(\x05\x12\x19\n\x11twoFactorRequired\x18\x13 \x01(\x08\x12\x16\n\x0e\x64isallowExport\x18\x14 \x01(\x08\x12\x15\n\rrestrictFiles\x18\x15 \x01(\x08\x12\x1a\n\x12restrictAllSharing\x18\x16 \x01(\x08\x12\x17\n\x0frestrictSharing\x18\x17 \x01(\x08\x12\"\n\x1arestrictSharingIncomingAll\x18\x18 \x01(\x08\x12)\n!restrictSharingIncomingEnterprise\x18\x19 \x01(\x08\x12\x13\n\x0blogoutTimer\x18\x1a \x01(\x03\x12\x17\n\x0fpersistentLogin\x18\x1b \x01(\x08\x12\x1c\n\x14ipDisableAutoApprove\x18\x1c \x01(\x08\x12$\n\x1cshareDataKeyWithEccPublicKey\x18\x1d \x01(\x08\x12\'\n\x1fshareDataKeyWithDevicePublicKey\x18\x1e \x01(\x08\x12\x1a\n\x12RecordTypesCounter\x18\x1f \x01(\x05\x12$\n\x1cRecordTypesEnterpriseCounter\x18  \x01(\x05\x12\x1a\n\x12recordTypesEnabled\x18! \x01(\x08\x12\x1c\n\x14\x63\x61nManageRecordTypes\x18\" \x01(\x08\x12\x1d\n\x15recordTypesPAMCounter\x18# \x01(\x05\x12\x1a\n\x12logoutTimerMinutes\x18$ \x01(\x05\x12 \n\x18securityKeysNoUserVerify\x18% \x01(\x08\x12\x36\n\x08\x63hannels\x18& \x03(\x0e\x32$.Authentication.TwoFactorChannelType\x12\x19\n\x11personalUsernames\x18\' \x03(\t\"&\n\x08KeyValue\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"-\n\x0fKeyValueBoolean\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x08\"*\n\x0cKeyValueLong\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x03\"=\n\x06Result\x12\x12\n\nresultCode\x18\x01 \x01(\t\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x0e\n\x06result\x18\x03 \x01(\t\"\xc2\x01\n\x0c\x45nforcements\x12)\n\x07strings\x18\x01 \x03(\x0b\x32\x18.AccountSummary.KeyValue\x12\x31\n\x08\x62ooleans\x18\x02 \x03(\x0b\x32\x1f.AccountSummary.KeyValueBoolean\x12+\n\x05longs\x18\x03 \x03(\x0b\x32\x1c.AccountSummary.KeyValueLong\x12\'\n\x05jsons\x18\x04 \x03(\x0b\x32\x18.AccountSummary.KeyValue\"<\n\x16MissingAccountShareKey\x12\x0f\n\x07role_id\x18\x01 \x01(\x03\x12\x11\n\tpublicKey\x18\x02 \x01(\x0c\"u\n\x0cPasswordRule\x12\x10\n\x08ruleType\x18\x01 \x01(\t\x12\x0f\n\x07pattern\x18\x02 \x01(\t\x12\r\n\x05match\x18\x03 \x01(\x08\x12\x0f\n\x07minimum\x18\x04 \x01(\x05\x12\x13\n\x0b\x64\x65scription\x18\x05 \x01(\t\x12\r\n\x05value\x18\x06 \x01(\t\"\x97\x01\n\x0bSecurityKey\x12\x10\n\x08\x64\x65viceId\x18\x01 \x01(\x03\x12\x12\n\ndeviceName\x18\x02 \x01(\t\x12\x11\n\tdateAdded\x18\x03 \x01(\x03\x12\x0f\n\x07isValid\x18\x04 \x01(\x08\x12>\n\x12\x64\x65viceRegistration\x18\x05 \x01(\x0b\x32\".AccountSummary.DeviceRegistration\"y\n\x12\x44\x65viceRegistration\x12\x11\n\tkeyHandle\x18\x01 \x01(\t\x12\x11\n\tpublicKey\x18\x02 \x01(\x0c\x12\x17\n\x0f\x61ttestationCert\x18\x03 \x01(\t\x12\x0f\n\x07\x63ounter\x18\x04 \x01(\x03\x12\x13\n\x0b\x63ompromised\x18\x05 \x01(\x08\"k\n\x05Group\x12\r\n\x05\x61\x64min\x18\x01 \x01(\x08\x12\x1d\n\x15groupVerificationCode\x18\x02 \x01(\t\x12\x34\n\radministrator\x18\x04 \x01(\x0b\x32\x1d.AccountSummary.Administrator\"\xc0\x01\n\rAdministrator\x12\x11\n\tfirstName\x18\x01 \x01(\t\x12\x10\n\x08lastName\x18\x02 \x01(\t\x12\r\n\x05\x65mail\x18\x03 \x01(\t\x12\x1c\n\x14\x63urrentNumberOfUsers\x18\x04 \x01(\x05\x12\x15\n\rnumberOfUsers\x18\x05 \x01(\x05\x12\x18\n\x10subscriptionCode\x18\x07 \x01(\t\x12\x16\n\x0e\x65xpirationDate\x18\x08 \x01(\t\x12\x14\n\x0cpurchaseDate\x18\t \x01(\tB*\n\x18\x63om.keepersecurity.protoB\x0e\x41\x63\x63ountSummaryb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'AccountSummary_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\030com.keepersecurity.protoB\016AccountSummary'
  _globals['_ACCOUNTSUMMARYREQUEST']._serialized_start=58
  _globals['_ACCOUNTSUMMARYREQUEST']._serialized_end=136
  _globals['_ACCOUNTSUMMARYELEMENTS']._serialized_start=139
  _globals['_ACCOUNTSUMMARYELEMENTS']._serialized_end=803
  _globals['_DEVICEINFO']._serialized_start=806
  _globals['_DEVICEINFO']._serialized_end=1201
  _globals['_KEYSINFO']._serialized_start=1204
  _globals['_KEYSINFO']._serialized_end=1397
  _globals['_SYNCLOG']._serialized_start=1400
  _globals['_SYNCLOG']._serialized_end=1529
  _globals['_LICENSE']._serialized_start=1532
  _globals['_LICENSE']._serialized_end=2403
  _globals['_ADDON']._serialized_start=2406
  _globals['_ADDON']._serialized_end=2569
  _globals['_SETTINGS']._serialized_start=2572
  _globals['_SETTINGS']._serialized_end=3785
  _globals['_KEYVALUE']._serialized_start=3787
  _globals['_KEYVALUE']._serialized_end=3825
  _globals['_KEYVALUEBOOLEAN']._serialized_start=3827
  _globals['_KEYVALUEBOOLEAN']._serialized_end=3872
  _globals['_KEYVALUELONG']._serialized_start=3874
  _globals['_KEYVALUELONG']._serialized_end=3916
  _globals['_RESULT']._serialized_start=3918
  _globals['_RESULT']._serialized_end=3979
  _globals['_ENFORCEMENTS']._serialized_start=3982
  _globals['_ENFORCEMENTS']._serialized_end=4176
  _globals['_MISSINGACCOUNTSHAREKEY']._serialized_start=4178
  _globals['_MISSINGACCOUNTSHAREKEY']._serialized_end=4238
  _globals['_PASSWORDRULE']._serialized_start=4240
  _globals['_PASSWORDRULE']._serialized_end=4357
  _globals['_SECURITYKEY']._serialized_start=4360
  _globals['_SECURITYKEY']._serialized_end=4511
  _globals['_DEVICEREGISTRATION']._serialized_start=4513
  _globals['_DEVICEREGISTRATION']._serialized_end=4634
  _globals['_GROUP']._serialized_start=4636
  _globals['_GROUP']._serialized_end=4743
  _globals['_ADMINISTRATOR']._serialized_start=4746
  _globals['_ADMINISTRATOR']._serialized_end=4938
# @@protoc_insertion_point(module_scope)



from __future__ import annotations
from .dag import DAG
from .edge import EdgeType
from .exceptions import DAGVertexException, DAGDataException
from .crypto import decrypt_aes, str_to_bytes


class Maintenance:

    def __init__(self, dag: DAG, sync_point: int = 0, decrypt=False):
        self.dag = dag
        self.sync_point = sync_point
        self.decrypt = decrypt
        self._loaded = False
        self._load_sync_point = None

    def debug(self, msg: str, level: int = 0):
        return self.dag.debug(msg, level)

    @property
    def logger(self):
        return self.dag.logger

    def _get_keychain(self, v):
        self.debug(f"getting keychain for vertex {v.uid}, {v.name}, {v.vertex_type}", level=1)

        keychain = []

        found_key_edge = False
        for e in v.edges:
            if e.edge_type == EdgeType.KEY:
                head = self.dag.get_vertex(e.head_uid)
                keychain += self._get_keychain(head)

                # Each vertex has a "keychain".
                # However, this will be one key in an array.
                try:
                    content = decrypt_aes(e.content, keychain[-1]["key"])
                    keychain += [
                        {
                            "uid": v.uid,
                            "name": v.name,
                            "type": v.vertex_type,
                            "key": content,
                            "corrupt": False
                        }
                    ]
                    found_key_edge = True
                    break
                except Exception as err:
                    self.logger.debug(f"could not decrypt key for {v.uid}, {keychain[-1]['key']}: {err}", level=1)  # DAG_DEBUG_LEVEL env var
                    keychain += [
                        {
                            "uid": v.uid,
                            "name": v.name,
                            "type": v.vertex_type,
                            "corrupt": True
                        }
                    ]
                    return keychain

        if found_key_edge is True:
            return keychain
        else:
            return [
                {
                    "uid": self.dag.uid,
                    "key": self.dag.key,
                    "name": self.dag.name,
                    "type": self.dag.vertex_type,
                    "corrupt": False
                }
            ]

    def load(self, sync_point: int = 0):
        if self._loaded is False or sync_point != self._load_sync_point:
            self.logger.info(f"reloading the graph with decrypt {self.decrypt} and sync point {self.sync_point}")
            # Disable the automatic decrypt of the KEY and DATA edges.
            self.dag.decrypt = self.decrypt
            self.dag.load(sync_point=self.sync_point)
            self._loaded = True
            self._load_sync_point = sync_point

    def reload(self):
        self._loaded = False
        self.load()

    def get_keychain(self, uid: str, sync_point: int = 0):
        """

        :param uid: Either the UID, or name of the vertex.
        :param sync_point: A starting sync point for loading the graph.
        :return:
        """

        # Disable the automatic decrypt of the KEY and DATA edges.
        self.dag.decrypt = False
        self.load(sync_point=sync_point)

        vertex = self.dag.get_vertex(uid)
        if vertex is None:
            raise DAGVertexException(f"Vertex {uid} does not exists.", uid=uid)

        key_chain = self._get_keychain(vertex)

        return key_chain

    def get_data(self, uid: str, key: bytes, sync_point: int = 0):

        self.dag.decrypt = False
        self.load(sync_point=sync_point)

        vertex = self.dag.get_vertex(uid)
        if vertex is None:
            raise DAGVertexException(f"Vertex {uid} does not exists.", uid=uid)

        content = vertex.content
        if content is None:
            raise DAGVertexException(f"Vertex {uid} does not have a DATA edge.", uid=uid)
        if isinstance(content, str) is False:
            raise DAGDataException(f"Vertex {uid} DATA edge content was a not str.", uid=uid)

        try:
            self.debug(f"decrypt {uid} data {content} with {key}")
            return decrypt_aes(str_to_bytes(content), key)
        except Exception as err:
            raise DAGDataException(f"Could not decrypt vertex {uid}'s data edge: {err}", uid=uid)

    def delete_data(self):
        pass



#  _  __  
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|            
#
# Keeper Commander 
# Contact: ops@keepersecurity.com
#

class Error(Exception):
    """Base class for exceptions in this module."""
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class KeeperApiError(Error):
    """Exception raised with failed Keeper API request
    """

    def __init__(self, result_code, message):
        super().__init__(message)
        self.result_code = result_code

    def __str__(self):
        return f'{self.result_code or ""}: {self.message or ""}'


class CommandError(Error):
    def __init__(self, command, message):
        super().__init__(message)
        self.command = command

    def __str__(self):
        if self.command:
            return f'{self.command}: {self.message}'
        else:
            return super().__str__()
        


#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2023 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import argparse
import copy
import datetime
import json
import logging
import os.path
import secrets
from typing import Optional, List, Tuple, Union

from . import aram
from .base import GroupCommand, report_output_parser, field_to_title, dump_report_data, raise_parse_exception, suppress_exit
from .enterprise_common import EnterpriseCommand
from .. import api, utils
from ..params import KeeperParams
from ..error import CommandError


PUT_ENTERPRISE_SETTING = 'put_enterprise_setting'
AUDIT_ALERT_CONTEXT = 'AuditAlertContext'

alert_list_parser = argparse.ArgumentParser(prog='audit-alert list', parents=[report_output_parser])
alert_list_parser.add_argument('--reload', dest='reload', action='store_true', help='reload alert information')

alert_target_parser = argparse.ArgumentParser(add_help=False)
alert_target_parser.add_argument('target', metavar='ALERT', help='Alert ID or Name.')
alert_view_parser = argparse.ArgumentParser(prog='audit-alert view')
alert_view_parser.add_argument('target', metavar='ALERT', nargs='?', help='Alert ID or Name')
alert_view_parser.add_argument('--all', dest='all', action='store_true', help='View all alerts')
alert_view_parser.add_argument('--format', dest='format', action='store', choices=['table', 'csv', 'json', 'pdf'],
                                  default='table', help='format of output')
alert_view_parser.add_argument('--output', dest='output', action='store',
                                  help='path to resulting output file (ignored for "table" format)')
alert_history_parser = argparse.ArgumentParser(
    prog='audit-alert history', parents=[report_output_parser, alert_target_parser])
alert_reset_counts_parser = argparse.ArgumentParser(prog='audit-alert reset-counts', parents=[alert_target_parser])
alert_delete_parser = argparse.ArgumentParser(prog='audit-alert remove')
alert_delete_parser.add_argument('target', metavar='ALERT', nargs='?', help='Alert ID or Name.')
alert_delete_parser.add_argument('--all', dest='delete_all', action='store_true', help='Delete all alerts')
alert_delete_parser.add_argument('--from', dest='from_id', metavar='ALERT ID', type=int,
                                help='Starting alert ID for range deletion')
alert_delete_parser.add_argument('--to', dest='to_id', metavar='ALERT ID', type=int,
                                help='Ending alert ID for range deletion')
alert_delete_parser.add_argument('--force', dest='force', action='store_true', 
                                help='Force deletion without confirmation prompt')

alert_recipient_edit_options = argparse.ArgumentParser(add_help=False)
alert_recipient_edit_options.add_argument(
    '--name', dest='name', metavar='NAME', action='store', help='recipient name')
alert_recipient_edit_options.add_argument(
    '--email', dest='email', metavar='EMAIL', action='store', help='email address')
alert_recipient_edit_options.add_argument(
    '--phone', dest='phone', metavar='PHONE', action='store', help='phone number. +1 (555) 555-1234')
alert_recipient_edit_options.add_argument(
    '--webhook', dest='webhook', metavar='URL', action='store',
    help='Webhook URL. See https://docs.keeper.io/enterprise-guide/webhooks')
alert_recipient_edit_options.add_argument(
    '--http-body', dest='http_body', metavar='HTTP_BODY', action='store',
    help='Webhook HTTP Body')
alert_recipient_edit_options.add_argument(
    '--cert-errors', dest='cert_errors', action='store', choices=['ignore', 'enforce'],
    help='Webhook SSL Certificate errors')
alert_recipient_edit_options.add_argument(
    '--generate-token', dest='generate_token', action='store_true', help='Generate new access token')


alert_recipient_parser = argparse.ArgumentParser(prog='audit-alert recipient', parents=[alert_target_parser])
subparsers = alert_recipient_parser.add_subparsers(title='recipient actions', dest='action')
alert_recipient_enable_parser = subparsers.add_parser('enable', help='enables recipient')
alert_recipient_enable_parser.add_argument(
    'recipient', metavar='RECIPIENT', help='Recipient ID or Name. Use "*" for "User who generated event"')
alert_recipient_enable_parser.error = raise_parse_exception
alert_recipient_enable_parser.exit = suppress_exit

alert_recipient_disable_parser = subparsers.add_parser('disable', help='disables recipient')
alert_recipient_disable_parser.add_argument(
    'recipient', metavar='RECIPIENT', help='Recipient ID or Name. Use "*" for "User who generated event"')
alert_recipient_disable_parser.error = raise_parse_exception
alert_recipient_disable_parser.exit = suppress_exit

alert_recipient_delete_parser = subparsers.add_parser('delete', help='deletes recipient')
alert_recipient_delete_parser.add_argument('recipient', metavar='RECIPIENT', help='Recipient ID or Name.')
alert_recipient_delete_parser.error = raise_parse_exception
alert_recipient_delete_parser.exit = suppress_exit

alert_recipient_add_parser = subparsers.add_parser('add', help='adds recipient', parents=[alert_recipient_edit_options])
alert_recipient_add_parser.error = raise_parse_exception
alert_recipient_add_parser.exit = suppress_exit

alert_recipient_edit_parser = subparsers.add_parser('edit', help='edit recipient', parents=[alert_recipient_edit_options])
alert_recipient_edit_parser.add_argument('recipient', metavar='RECIPIENT', help='Recipient ID or Name.')
alert_recipient_edit_parser.error = raise_parse_exception
alert_recipient_edit_parser.exit = suppress_exit


alert_edit_options = argparse.ArgumentParser(add_help=False)
alert_edit_options.add_argument('--name', action='store', metavar='NAME', help='Alert Name.')
alert_edit_options.add_argument(
    '--frequency', dest='frequency', action='store', metavar='FREQUENCY',
    help='Alert Frequency. "[N:]event|minute|hour|day"')
alert_edit_options.add_argument(
    '--audit-event', dest='audit_event', action='append', metavar='EVENT',
    help='Audit Event. Can be repeated.')
alert_edit_options.add_argument(
    '--user', dest='user', action='append', metavar='USER', help='Username. Can be repeated.')
alert_edit_options.add_argument(
    '--record-uid', dest='record_uid', action='append', metavar='RECORD_UID', help='Record UID. Can be repeated.')
alert_edit_options.add_argument(
    '--shared-folder-uid', dest='shared_folder_uid', action='append', metavar='SHARED_FOLDER_UID',
    help='Shared Folder UID. Can be repeated.')
alert_edit_options.add_argument(
    '--active', dest='active', action='store', metavar='ACTIVE', choices=['on', 'off'],
    help='Enable or disable alert')


alert_add_parser = argparse.ArgumentParser(prog='audit-alert add', parents=[alert_edit_options])
alert_edit_parser = argparse.ArgumentParser(prog='audit-alert edit', parents=[alert_target_parser, alert_edit_options])

alert_action_parser = argparse.ArgumentParser(prog='audit-alert action')
alert_action_parser.add_argument('target', metavar='ALERT', nargs='?', help='Alert ID or Name')
alert_action_parser.add_argument('--all', dest='apply_all', action='store_true', help='Apply action to all alerts')
alert_action_parser.error = raise_parse_exception
alert_action_parser.exit = suppress_exit


class AuditSettingMixin:
    LAST_USERNAME = ""
    LAST_ENTERPRISE_ID = 0
    SETTINGS = None   # type: Optional[dict]
    EVENT_TYPES = None  # type: Optional[List[Tuple[int, str]]]

    @staticmethod
    def load_settings(params, reload=False):  # type: (KeeperParams, Optional[bool]) -> Optional[dict]
        if params.enterprise is None:
            AuditSettingMixin.SETTINGS = None
            AuditSettingMixin.LAST_ENTERPRISE_ID = 0
            return
        if AuditSettingMixin.EVENT_TYPES is None:
            rq = {
                'command': 'get_audit_event_dimensions',
                'columns': ['audit_event_type']
            }
            rs = api.communicate(params, rq)
            AuditSettingMixin.EVENT_TYPES = []
            for et in rs['dimensions']['audit_event_type']:
                event_id = et.get('id')
                event_name = et.get('name')
                if event_name and isinstance(event_id, int):
                    AuditSettingMixin.EVENT_TYPES.append((event_id, event_name))

        enterprise_id = 0
        if isinstance(params.license, dict):
            enterprise_id = params.license.get('enterprise_id')

        if AuditSettingMixin.SETTINGS is None:
            reload = True
        elif AuditSettingMixin.LAST_USERNAME != params.user:
            reload = True
        elif enterprise_id == 0 or AuditSettingMixin.LAST_ENTERPRISE_ID != enterprise_id:
            reload = True

        if reload:
            rq = {
                'command': 'get_enterprise_setting',
                'include': ['AuditAlertContext', 'AuditAlertFilter', 'AuditReportFilter']
            }
            AuditSettingMixin.SETTINGS = api.communicate(params, rq)
            AuditSettingMixin.LAST_USERNAME = params.user
            AuditSettingMixin.LAST_ENTERPRISE_ID = enterprise_id
        return AuditSettingMixin.SETTINGS

    @staticmethod
    def invalidate_alerts():
        AuditSettingMixin.SETTINGS = None

    @staticmethod
    def frequency_to_text(freq):   # type: (dict) -> Optional[str]
        if not isinstance(freq, dict):
            return
        period = freq.get('period')
        count = freq.get('count')
        if period == 'event':
            if isinstance(count, int):
                return f'{count} of Occurrences Triggered'
            else:
                return 'Every Occurrence'
        elif period in ('day', 'hour', 'minutes') and isinstance(count, int):
            if period == 'minutes':
                period = 'minute'
            period = period.capitalize()
            return f'{count} {period}(s) from First Occurrence'
        else:
            return 'Not supported'

    @staticmethod
    def text_to_frequency(text):   # type: (str) -> dict
        if not isinstance(text, str):
            return {'period': 'event'}
        num, sep, occ = text.partition(':')
        if sep:
            if num:
                num = int(num)
            else:
                num = 0
            occ = occ.lower()
        else:
            num = 0
            occ = text.lower()
        if occ in ('event', 'e'):
            occ = 'event'
        elif occ in ('minute', 'minutes', 'm'):
            occ = 'minutes'
        elif occ in ('hour', 'h'):
            occ = 'hour'
        elif occ in ('day', 'd'):
            occ = 'day'
        else:
            raise Exception(f'Invalid alert frequency \"{occ}\". "event", "day", "hour", "minute"')
        if num <= 0:
            if occ == 'event':
                num = 0
            else:
                num = 1
        freq = {
            'period': occ
        }
        if num > 0:
            freq['count'] = num
        return freq

    @staticmethod
    def get_alert_context(alert_id):   # type: (int) -> Optional[dict]
        settings = AuditSettingMixin.SETTINGS
        if not settings:
            return
        alert_context = settings.get('AuditAlertContext')
        if not isinstance(alert_context, list):
            return

        return next((x for x in alert_context if x.get('id') == alert_id), None)

    @staticmethod
    def get_alert_configuration(params, alert_name):   # type: (KeeperParams, str) -> dict
        if not alert_name:
            raise Exception(f'Alert name cannot be empty')

        settings = AuditSettingMixin.load_settings(params)
        if not settings:
            raise Exception(f'Alert with name \"{alert_name}\" not found')
        alert_filter = settings.get('AuditAlertFilter')
        if not isinstance(alert_filter, list):
            raise Exception(f'Alert with name \"{alert_name}\" not found')

        a_number = int(alert_name) if alert_name.isnumeric() else 0
        if a_number > 0:
            for alert_filter in alert_filter:
                a_id = alert_filter.get('id')
                if isinstance(a_id, int):
                    if a_id == a_number:
                        return alert_filter

        alerts = []
        l_name = alert_name.casefold()
        for alert_filter in alert_filter:
            a_name = alert_filter.get('name') or ''
            if a_name.casefold() == l_name:
                alerts.append(alert_filter)

        if len(alerts) == 0:
            raise Exception(f'Alert with name \"{alert_name}\" not found')
        if len(alerts) > 1:
            raise Exception(f'There are {len(alerts)} alerts with name \"{alert_name})\". Use alert ID.')
        return alerts[0]

    @staticmethod
    def apply_alert_options(params, alert, **kwargs):     # type: (KeeperParams, dict, ...) ->  None
        alert_name = kwargs.get('name')
        if alert_name:
            alert['name'] = alert_name

        frequency = kwargs.get('frequency')
        if frequency:
            alert['frequency'] = AuditSettingMixin.text_to_frequency(frequency)

        alert_filter = alert.get('filter')
        if not isinstance(alert_filter, dict):
            alert['filter'] = {}

        events_option = kwargs.get('audit_event')
        if isinstance(events_option, list):
            event_ids = set()
            event_lookup = {n: i for i, n in AuditSettingMixin.EVENT_TYPES or []}
            for events in events_option:    # type: str
                for event_name in (x.strip().lower() for x in events.split(',')):
                    if event_name in event_lookup:
                        event_ids.add(event_lookup[event_name])
                    else:
                        raise CommandError('alert add', f'Event name \"{event_name}\" is invalid')
            if len(event_ids) > 0:
                event_list = list(event_ids)
                event_list.sort()
                alert_filter['events'] = event_list
            else:
                if 'events' in alert_filter:
                    del alert_filter['event']

        users_option = kwargs.get('user')
        if isinstance(users_option, list):
            user_ids = set()
            user_lookup = {x['username']: x for x in params.enterprise.get('users') or []}
            # TODO aliases
            for users in users_option:    # type: str
                for username in (x.strip().lower() for x in users.split(',')):
                    if username in user_lookup:
                        user_ids.add(user_lookup[username]['user_id'])
                    else:
                        raise CommandError('alert add', f'Username \"{username}\" is unknown')
            if len(user_ids) > 0:
                alert_filter['userIds'] = list(user_ids)
            else:
                if 'userIds' in alert_filter:
                    del alert_filter['userIds']

        record_uid_option = kwargs.get('record_uid')
        if isinstance(record_uid_option, list):
            record_uids = set()
            for r_uids in record_uid_option:    # type: str
                for record_uid in (x.strip() for x in r_uids.split(',')):
                    if not record_uid:
                        continue
                    if record_uid not in params.record_cache:
                        logging.info('Record UID \"%s\" cannot be verified as existing.', record_uid)
                    record_uids.add(record_uid)
            if len(record_uids) > 0:
                alert_filter['recordUids'] = [{'id': x, 'selected': True} for x in record_uids]
            else:
                if 'recordUids' in alert_filter:
                    del alert_filter['recordUids']

        shared_folder_uid_option = kwargs.get('shared_folder_uid')
        if isinstance(shared_folder_uid_option, list):
            shared_folder_uids = set()
            for sf_uids in shared_folder_uid_option:    # type: str
                for shared_folder_uid in (x.strip() for x in sf_uids.split(',')):
                    if not shared_folder_uid:
                        continue
                    if shared_folder_uid not in params.shared_folder_cache:
                        logging.info('Shared Folder UID \"%s\" cannot be verified as existing.', shared_folder_uid)
                    shared_folder_uids.add(shared_folder_uid)
            if len(shared_folder_uids) > 0:
                alert_filter['sharedFolderUids'] = [{'id': x, 'selected': True} for x in shared_folder_uids]
            else:
                if 'sharedFolderUids' in alert_filter:
                    del alert_filter['sharedFolderUids']


class AuditAlertList(EnterpriseCommand, AuditSettingMixin):
    def get_parser(self):
        return alert_list_parser

    def execute(self, params, **kwargs):
        alerts = self.load_settings(params, kwargs.get('reload') or False)
        if not isinstance(alerts, dict):
            logging.info('No alerts found')
            return
        alert_filter = alerts.get('AuditAlertFilter')
        if not isinstance(alert_filter, list):
            logging.info('No alerts found')
            return

        fmt = kwargs.get('format') or ''
        table = []
        headers = ['id', 'name', 'events', 'frequency', 'occurrences', 'alerts_sent', 'last_sent', 'active']
        if fmt != 'json':
            headers = [field_to_title(x) for x in headers]
        event_lookup = {i: n for i, n in self.EVENT_TYPES or []}
        for alert in alert_filter:
            alert_id = alert.get('id')
            ctx = AuditSettingMixin.get_alert_context(alert_id) or alert
            alert_name = alert.get('name')
            events = ''   # type: Union[str, List[str]]
            alert_filter = alert.get('filter')
            if isinstance(alert_filter, dict):
                es = list((event_lookup[x] for x in alert_filter.get('events') or [] if x in event_lookup))
                if len(es) == 1:
                    events = es[0]
                elif len(es) <= 5:
                    events = '\n'.join(es)
                elif len(es) > 5:
                    events = '\n'.join(es[:4]) + f'\n+{len(es) - 4} more'
            freq = self.frequency_to_text(alert.get('frequency'))
            occurrences = ctx.get('counter')
            alerts_sent = ctx.get('sentCounter')
            last_sent = ctx.get('lastSent')
            if last_sent:
                try:
                    last_sent = datetime.datetime.strptime(last_sent, '%Y-%m-%dT%H:%M:%S.%fZ')
                    last_sent = last_sent.replace(microsecond=0, tzinfo=datetime.timezone.utc).astimezone()
                except ValueError:
                    pass
            disabled = ctx.get('disabled') is True
            table.append([alert_id, alert_name, events, freq, occurrences, alerts_sent, last_sent, not disabled])
        return dump_report_data(table, headers, fmt=fmt, filename=kwargs.get('output'), sort_by=0)


class AuditAlertDelete(EnterpriseCommand, AuditSettingMixin):
    def get_parser(self):
        return alert_delete_parser

    def execute(self, params, **kwargs):
        target = kwargs.get('target')
        delete_all = kwargs.get('delete_all')
        from_id = kwargs.get('from_id')
        to_id = kwargs.get('to_id')
        
        # Load settings to get all alerts
        settings = self.load_settings(params)
        if not settings:
            logging.info('No alerts found')
            return
        alert_filter = settings.get('AuditAlertFilter')
        if not isinstance(alert_filter, list) or len(alert_filter) == 0:
            logging.info('No alerts found')
            return

        # Determine which deletion method to use and get alerts to delete
        if from_id is not None or to_id is not None:
            alerts_to_delete = self._delete_by_range(params, from_id, to_id, alert_filter)
        elif delete_all:
            alerts_to_delete = self._delete_all(params, alert_filter)
        elif target:
            alerts_to_delete = self._delete_single(params, target)
        else:
            raise CommandError('alert delete', 'Either target, --all, or --from/--to parameters are required.')

        # Confirm deletion unless --force is used
        force = kwargs.get('force', False)
        if not force and not self._confirm_deletion(alerts_to_delete):
            logging.info('Deletion cancelled by user.')
            return

        # Execute the deletion
        deleted_count = self._execute_deletion(params, alerts_to_delete)
        
        self.invalidate_alerts()
        
        if deleted_count > 0:
            # Show remaining alerts
            command = AuditAlertList()
            command.execute(params, reload=True)
        else:
            logging.warning('No alerts were deleted.')

    def _delete_by_range(self, params, from_id, to_id, alert_filter):
        """Handle range deletion with --from and --to parameters."""
        if from_id is None or to_id is None:
            raise CommandError('alert delete', 'Both --from and --to parameters are required for range deletion.')
        
        # Validation: both values should not be zero or negative
        if from_id <= 0 or to_id <= 0:
            raise CommandError('alert delete', 'Alert IDs must be positive integers. Please specify valid alert IDs.')
        
        # Validation: from value should be less than to value
        if from_id >= to_id:
            raise CommandError('alert delete', f'Invalid range: --from ({from_id}) must be less than --to ({to_id}).')
        
        # Find alerts in range
        alerts_to_delete = []
        for alert in alert_filter:
            alert_id = alert.get('id')
            if isinstance(alert_id, int) and from_id <= alert_id <= to_id:
                alerts_to_delete.append(alert)
        
        if not alerts_to_delete:
            raise CommandError('alert delete', f'No alerts found in range {from_id}-{to_id}')
        
        return alerts_to_delete

    def _delete_all(self, params, alert_filter):
        """Handle delete all with --all flag."""
        return alert_filter

    def _delete_single(self, params, target):
        """Handle single target deletion."""
        alert = AuditSettingMixin.get_alert_configuration(params, target)
        return [alert]

    def _confirm_deletion(self, alerts_to_delete):
        """Prompt user for confirmation before deleting alerts."""
        if not alerts_to_delete:
            return False
        
        print(f"\nThe following {len(alerts_to_delete)} alert(s) will be deleted:")
        print("-" * 60)
        for alert in alerts_to_delete:
            alert_id = alert.get('id', 'N/A')
            alert_name = alert.get('name', 'N/A')
            print(f"  ID: {alert_id} | Name: {alert_name}")
        print("-" * 60)
        
        try:
            response = input(f"Are you sure you want to delete {len(alerts_to_delete)} alert(s)? (y/n): ").strip().lower()
            return response in ('y', 'yes')
        except (KeyboardInterrupt, EOFError):
            print("\nOperation cancelled.")
            return False

    def _execute_deletion(self, params, alerts_to_delete):
        """Execute the actual deletion of alerts and return count of successfully deleted alerts."""
        deleted_count = 0
        for alert in alerts_to_delete:
            alert_id = alert.get('id')
            alert_name = alert.get('name', f'ID {alert_id}')
            try:
                rq = {
                    'command': 'delete_enterprise_setting',
                    'type': 'AuditAlertFilter',
                    'id': alert_id,
                }
                api.communicate(params, rq)
                deleted_count += 1
            except Exception as e:
                logging.error(f'Failed to delete alert: {alert_name} (ID: {alert_id}): {str(e)}')
        
        return deleted_count


class AuditAlertView(EnterpriseCommand, AuditSettingMixin):
    def get_parser(self):
        return alert_view_parser

    def execute(self, params, **kwargs):
        fmt = kwargs.get('format', '')
        view_all = kwargs.get('all') or not kwargs.get('target')
        
        if view_all:
            self._execute_view_all(params, kwargs, fmt)
        else:
            self._execute_view_single(params, kwargs, fmt)

    def _get_display_options(self, kwargs):
        """Determine what sections to show based on kwargs."""
        if kwargs.get('recipient_only'):
            return {'recipient': True, 'filter': False, 'stat': False}
        return {'recipient': True, 'filter': True, 'stat': True}

    def _get_headers(self, show_options, fmt):
        """Get appropriate headers based on display options."""
        if show_options['recipient'] and not show_options['filter'] and not show_options['stat']:
            headers = ['alert_id', 'alert_name', 'status', 'recipients']
        elif show_options['stat']:
            headers = ['alert_id', 'alert_name', 'status', 'frequency', 'occurrences', 
                      'sent_counter', 'last_sent', 'event_types', 'users', 'shared_folders', 
                      'records', 'recipients']
        else:
            headers = ['alert_id', 'alert_name', 'status', 'event_types', 'users', 
                      'shared_folders', 'records', 'recipients']
        
        return headers if fmt == 'json' else [field_to_title(x) for x in headers]

    def _format_datetime(self, date_str):
        """Format datetime string to local time."""
        if not date_str:
            return ''
        try:
            dt = datetime.datetime.strptime(date_str, '%Y-%m-%dT%H:%M:%S.%fZ')
            dt = dt.replace(microsecond=0, tzinfo=datetime.timezone.utc).astimezone()
            return dt.isoformat()
        except ValueError:
            return ''

    def _get_event_names(self, event_ids):
        """Convert event IDs to names."""
        if not event_ids:
            return []
        event_lookup = {i: n for i, n in self.EVENT_TYPES or []}
        return [event_lookup[x] for x in event_ids if x in event_lookup]

    def _get_usernames(self, user_ids, params):
        """Convert user IDs to usernames."""
        if not user_ids:
            return []
        user_lookup = {x['user_id']: x['username'] for x in params.enterprise.get('users', [])}
        return [user_lookup[x] for x in user_ids if x in user_lookup]

    def _get_selected_ids(self, items):
        """Get IDs of selected items."""
        if not items:
            return []
        return [x['id'] for x in items if x.get('selected')]

    def _build_recipient_data(self, recipient):
        """Build recipient data dictionary."""
        data = {
            "Recipient ID": recipient.get('id', ''),
            "Name": recipient.get('name', ''),
            "Status": 'Disabled' if recipient.get('disabled') else 'Enabled'
        }
        
        if 'webhook' in recipient:
            wh = recipient['webhook']
            data["Webhook URL"] = wh.get('url', '')
            if wh.get('template'):
                data["HTTP Body"] = wh.get('template', '')
            if wh.get('token'):
                data["Webhook Token"] = wh.get('token', '')
            data["Certificate Errors"] = 'Ignore' if wh.get('allowUnverifiedCertificate') else 'Enforce'
        
        if recipient.get('email'):
            data["Email To"] = recipient.get('email')
        
        if recipient.get('phone'):
            phone_country = recipient.get('phoneCountry')
            phone_display = f"(+{phone_country}) {recipient.get('phone')}" if phone_country else recipient.get('phone')
            data["Text To"] = phone_display
        
        return data

    def _format_list_for_cell(self, items, fmt, max_lines=5):
        """Format a list of items for display in a table cell, with line limits for PDF."""
        if not items:
            return ''
        
        if fmt == 'pdf' and len(items) > max_lines:
            displayed_items = items[:max_lines]
            additional_count = len(items) - max_lines
            return '\n'.join(displayed_items) + f'\n+ {additional_count} more'
        
        return '\n'.join(items)

    def _format_recipients_for_cell(self, recipients_complete, fmt):
        """Format recipients data for display in a table cell, with limits for PDF."""
        recipients_json = json.dumps(recipients_complete, indent=2)
        
        if fmt == 'pdf':
            lines = recipients_json.split('\n')
            if len(lines) > 5:
                truncated_lines = lines[:5]
                additional_count = len(lines) - 5
                return '\n'.join(truncated_lines) + f'\n+ {additional_count} more lines'
        
        return recipients_json

    def _build_alert_row(self, alert, show_options, params, fmt=''):
        """Build a single alert row for the table."""
        alert_id = alert.get('id', '')
        ctx = AuditSettingMixin.get_alert_context(alert_id) or alert
        alert_filter_data = alert.get('filter', {})
        
        row = [
            alert_id,
            alert.get('name', ''),
            'Disabled' if ctx.get('disabled') else 'Enabled'
        ]
        
        if show_options['stat'] and not show_options.get('recipient_only'):
            row.extend([
                self.frequency_to_text(alert.get('frequency')) or '',
                ctx.get('counter', ''),
                ctx.get('sentCounter', ''),
                self._format_datetime(ctx.get('lastSent'))
            ])
        
        if not show_options.get('recipient_only'):
            events = self._get_event_names(alert_filter_data.get('events'))
            row.append(self._format_list_for_cell(events, fmt))
            
            users = self._get_usernames(alert_filter_data.get('userIds'), params)
            row.append(self._format_list_for_cell(users, fmt))
            
            folders = self._get_selected_ids(alert_filter_data.get('sharedFolderUids'))
            row.append(self._format_list_for_cell(folders, fmt))
            
            records = self._get_selected_ids(alert_filter_data.get('recordUids'))
            row.append(self._format_list_for_cell(records, fmt))
        
        recipients_data = [self._build_recipient_data(r) for r in alert.get('recipients', [])]
        recipients_complete = {
            "Send To Originator": alert.get('sendToOriginator', False),
            "Recipients": recipients_data
        }
        row.append(self._format_recipients_for_cell(recipients_complete, fmt))
        
        return row

    def _execute_view_all(self, params, kwargs, fmt):
        """Execute view all alerts functionality."""
        show_options = self._get_display_options(kwargs)
        
        alerts = self.load_settings(params, kwargs.get('reload', False))
        if not isinstance(alerts, dict):
            logging.info('No alerts found')
            return
        
        alert_filter = alerts.get('AuditAlertFilter')
        if not isinstance(alert_filter, list):
            logging.info('No alerts found')
            return
        
        headers = self._get_headers(show_options, fmt)
        table = [self._build_alert_row(alert, show_options, params, fmt) for alert in alert_filter]
        
        dump_report_data(table, headers, fmt=fmt, filename=kwargs.get('output'), sort_by=None)

    def _add_filter_data_to_table(self, table, alert_filter, params):
        """Add filter data to table for single alert view."""
        table.extend([['', ''], ['Alert Filter:', '']])
        
        if 'events' in alert_filter:
            events = self._get_event_names(alert_filter.get('events'))
            table.append(['Event Types', events])
        
        if 'userIds' in alert_filter:
            users = self._get_usernames(alert_filter.get('userIds'), params)
            table.append(['User', users])
        
        if 'sharedFolderUids' in alert_filter:
            folders = self._get_selected_ids(alert_filter['sharedFolderUids'])
            table.append(['Shared Folder', folders])
        
        if 'recordUids' in alert_filter:
            records = self._get_selected_ids(alert_filter['recordUids'])
            table.append(['Record', records])

    def _add_recipient_data_to_table(self, table, alert):
        """Add recipient data to table for single alert view."""
        table.extend([['', ''], ['Recipients:', ''], 
                     ['Send To Originator (*)', alert.get('sendToOriginator', False)]])
        
        for recipient in alert.get('recipients', []):
            table.extend([
                ['', ''],
                ['Recipient ID', recipient.get('id')],
                ['Name', recipient.get('name')],
                ['Status', 'Disabled' if recipient.get('disabled') else 'Enabled']
            ])
            
            if 'webhook' in recipient:
                wh = recipient['webhook']
                table.append(['Webhook URL', wh.get('url')])
                if wh.get('template'):
                    table.append(['HTTP Body', wh.get('template')])
                table.append(['Webhook Token', wh.get('token')])
                table.append(['Certificate Errors', 'Ignore' if wh.get('allowUnverifiedCertificate') else 'Enforce'])
            
            if recipient.get('email'):
                table.append(['Email To', recipient.get('email')])
            
            if recipient.get('phone'):
                phone_country = recipient.get('phoneCountry')
                phone_display = (f'(+{phone_country}) {recipient.get("phone")}' 
                               if phone_country else recipient.get('phone'))
                table.append(['Text To', phone_display])

    def _execute_view_single(self, params, kwargs, fmt):
        """Execute view single alert functionality."""
        show_options = self._get_display_options(kwargs)
        target = kwargs.get('target')
        
        alert = AuditSettingMixin.get_alert_configuration(params, target)
        alert_id = alert.get('id')
        ctx = AuditSettingMixin.get_alert_context(alert_id) or alert
        
        header = ['name', 'value'] if fmt == 'json' else [field_to_title(x) for x in ['name', 'value']]
        
        table = [
            ['Alert ID', alert_id],
            ['Alert name', alert.get('name')],
            ['Status', 'Disabled' if ctx.get('disabled') else 'Enabled']
        ]
        
        if show_options['stat']:
            last_sent = ctx.get('lastSent')
            if last_sent:
                try:
                    last_sent = datetime.datetime.strptime(last_sent, '%Y-%m-%dT%H:%M:%S.%fZ')
                    last_sent = last_sent.replace(microsecond=0, tzinfo=datetime.timezone.utc).astimezone()
                except ValueError:
                    pass
            
            table.extend([
                ['Frequency', self.frequency_to_text(alert.get('frequency'))],
                ['Occurrences', ctx.get('counter')],
                ['Sent Counter', ctx.get('sentCounter')],
                ['Last Sent', last_sent.isoformat() if last_sent else '']
            ])
        
        if show_options['filter']:
            alert_filter = alert.get('filter', {})
            self._add_filter_data_to_table(table, alert_filter, params)
        
        if show_options['recipient']:
            self._add_recipient_data_to_table(table, alert)
        
        dump_report_data(table, header, no_header=True, right_align=(0,), 
                        fmt=fmt, filename=kwargs.get('output'), sort_by=None)

class AuditAlertHistory(EnterpriseCommand, AuditSettingMixin):
    def get_parser(self):
        return alert_history_parser

    def execute(self, params, **kwargs):
        alert = AuditSettingMixin.get_alert_configuration(params, kwargs.get('target'))
        command = aram.AuditReportCommand()
        json_str = command.execute(params, report_type='raw', report_format='fields',
                                   alert_uid=alert.get('alertUid'), event_type='audit_alert_sent',
                                   limit=100, order='desc', format='json')
        events = json.loads(json_str)
        fmt = kwargs.get('format') or ''
        headers = ['alert_sent_at', 'occurrences']
        if fmt != 'json':
            headers = [field_to_title(x) for x in headers]
        table = []
        for event in events:
            if 'recipient' in event:
                recipient = event.get('recipient')
                if recipient == 'throttled':
                    if len(table) > 0:
                        table[-1][1] += 1
                else:
                    table.append([event.get('created'), 1])
        return dump_report_data(table, headers, fmt=fmt, filename=kwargs.get('output'))


class AuditAlertResetCount(EnterpriseCommand, AuditSettingMixin):
    def get_parser(self):
        return alert_reset_counts_parser

    def execute(self, params, **kwargs):
        alert = AuditSettingMixin.get_alert_configuration(params, kwargs.get('target'))
        rq = {
            'command': 'put_enterprise_setting',
            'type': 'AuditAlertContext',
            'settings': {
                'id': alert.get('id'),
                'counter': 0,
                'sentCounter': 0,
                'lastReset': utils.current_milli_time()
            }
        }
        api.communicate(params, rq)
        AuditSettingMixin.invalidate_alerts()
        logging.info('Alert counts reset to zero')

class AuditAlertSent(EnterpriseCommand, AuditSettingMixin):    
    def __init__(self, action=None):
        super().__init__()
        self.action = action
    
    def get_parser(self):
        return alert_action_parser

    def execute(self, params, **kwargs):
        """Execute enable or disable action based on the command used"""
        apply_all = kwargs.get('apply_all', False)
        target = kwargs.get('target')
        action = self.action
        command_name = f'alert {action}'
        
        if apply_all and target:
            raise CommandError(command_name, 'Cannot specify both alert target and --all flag')
        elif apply_all:
            self._apply_to_all_alerts(params, action)
        elif not target:
            raise CommandError(command_name, 'Alert ID/Name is required unless using --all flag')
        else:
            self._apply_to_single_alert(params, target, action)
    
    def _apply_to_single_alert(self, params, target, action):
        alert = AuditSettingMixin.get_alert_configuration(params, target)
        disabled_value = action == 'disable'

        rq = {
            'command': PUT_ENTERPRISE_SETTING,
            'type': AUDIT_ALERT_CONTEXT,
            'settings': {
                'id': alert.get('id'),
                'disabled': disabled_value
            }
        }
        api.communicate(params, rq)
        AuditSettingMixin.invalidate_alerts()
        
        alert_name = alert.get('name') or f"Alert ID {alert.get('id')}"
        action_past = 'enabled' if action == 'enable' else 'disabled'
        logging.info(f'Alert "{alert_name}" has been {action_past}')
        command = AuditAlertView()
        command.execute(params, target=target)
    
    def _apply_to_all_alerts(self, params, action):
        alerts = self.load_settings(params)
        if not isinstance(alerts, dict):
            logging.info('No alerts found')
            return
        alert_filter = alerts.get('AuditAlertFilter')
        if not isinstance(alert_filter, list):
            logging.info('No alerts found')
            return
        
        disabled_value = action == 'disable'
        requests = []
        for alert in alert_filter:
            alert_id = alert.get('id')
            if alert_id:
                rq = {
                    'command': PUT_ENTERPRISE_SETTING,
                    'type': AUDIT_ALERT_CONTEXT,
                    'settings': {
                        'id': alert_id,
                        'disabled': disabled_value
                    }
                }
                requests.append(rq)
        
        if not requests:
            action_verb = 'enable' if action == 'enable' else 'disable'
            logging.info(f'No valid alerts found to {action_verb}')
            return
        
        api.execute_batch(params, requests)
        action_past = 'Enabled' if action == 'enable' else 'Disabled'
        logging.info(f'{action_past} {len(requests)} alert(s)')
  
        AuditSettingMixin.invalidate_alerts()
        command = AuditAlertList()
        command.execute(params)


class AuditAlertRecipients(EnterpriseCommand, AuditSettingMixin):
    def get_parser(self):
        return alert_recipient_parser

    def execute(self, params, **kwargs):
        alert = AuditSettingMixin.get_alert_configuration(params, kwargs.get('target'))
        alert = copy.deepcopy(alert)
        action = kwargs.get('action')
        skip_update = False
        if action in ('enable', 'disable'):
            name = kwargs.get('recipient')
            if name == '*':
                alert['sendToOriginator'] = action == 'enable'
            else:
                r = self.find_recipient(alert, name)
                r['disabled'] = action == 'disable'
        elif action == 'delete':
            r = self.find_recipient(alert, kwargs.get('recipient'))
            alert['recipients'].remove(r)
        elif action == 'edit':
            r = self.find_recipient(alert, kwargs.get('recipient'))
            self.apply_recipient(r, kwargs)
        elif action == 'add':
            if 'recipients' not in alert:
                alert['recipients'] = []
            ids = {x['id'] for x in alert['recipients'] if 'id' in x}
            r = {}
            for i in range(1000):
                if i+1 not in ids:
                    r['id'] = i + 1
                    break
            alert['recipients'].append(r)
            self.apply_recipient(r, kwargs)
        else:
            skip_update = True

        if not skip_update:
            rq = {
                'command': 'put_enterprise_setting',
                'type': 'AuditAlertFilter',
                'settings': alert,
            }
            api.communicate(params, rq)
            self.invalidate_alerts()
        command = AuditAlertView()
        command.execute(params, target=kwargs.get('target'), recipient_only=True)

    @staticmethod
    def apply_recipient(recipient, options):
        name = options.get('name')
        if name:
            recipient['name'] = name
        email = options.get('email')
        if email is not None:
            recipient['email'] = email
        phone = options.get('phone')
        if phone is not None:
            if phone:
                if phone.startswith('+'):
                    pc = ''
                    phone = phone[1:].strip()
                    while len(phone) > 0:
                        if phone[0:1].isnumeric():
                            pc += phone[0:1]
                            phone = phone[1:]
                        else:
                            break
                    phone_country = int(pc) if pc else 1
                    phone = phone.strip()
                else:
                    phone_country = 1
                recipient['phoneCountry'] = phone_country
                recipient['phone'] = phone
            else:
                recipient['phone'] = ''
                if 'phoneCountry' in recipient:
                    del recipient['phoneCountry']
        webhook = options.get('webhook')
        if webhook is not None:
            if webhook == '':
                if 'webhook' in recipient:
                    del recipient['webhook']
            else:
                if 'webhook' not in recipient:
                    recipient['webhook'] = {
                        'url': webhook,
                        'allowUnverifiedCertificate': False,
                        'token': utils.generate_uid()
                    }
                else:
                    recipient['webhook']['url'] = webhook
        http_body = options.get('http_body')
        if http_body is not None:
            if 'webhook' in recipient:
                webhook = recipient['webhook']
                if http_body:
                    if http_body[0] == '@':
                        file_name = http_body[1:]
                        file_name = os.path.expanduser(file_name)
                        if os.path.isfile(file_name):
                            with open(file_name, 'rt', encoding='utf-8') as tf:
                                webhook_body = tf.read()
                        else:
                            raise CommandError('', f'File \"{file_name}\" not found')
                        webhook['template'] = webhook_body
                    else:
                        webhook['template'] = http_body
                elif 'template' in webhook:
                    webhook['template'] = None

        cert_errors = options.get('cert_errors')
        if cert_errors is not None:
            if 'webhook' in recipient:
                recipient['webhook']['allowUnverifiedCertificate'] = cert_errors == 'ignore'
        if options.get('generate_token') is True:
            recipient['webhook']['token'] = utils.generate_uid()

    @staticmethod
    def find_recipient(alert, name):  # type: (dict, str) -> dict
        recs = []
        if isinstance(alert, dict):
            recipients = alert.get('recipients')
            if isinstance(recipients, list):
                r_id = int(name) if name.isnumeric() else -1
                if r_id > 0:
                    for r in recipients:
                        if r.get('id') == r_id:
                            return r
                l_name = name.lower()
                for r in recipients:
                    if (r.get('name') or '').lower() == l_name:
                        recs.append(r)
        if len(recs) == 0:
            raise Exception(f'Recipient \"{name}\" not found')
        if len(recs) > 1:
            raise Exception(f'There are {len(recs)} recipients with name \"{name}\". User recipient ID.')
        return recs[0]


class AuditAlertAdd(EnterpriseCommand, AuditSettingMixin):
    def get_parser(self):
        return alert_add_parser

    def execute(self, params, **kwargs):
        name = kwargs.get('name')
        if not name:
            raise CommandError('alert add', 'Alert name is required parameter')

        settings = self.load_settings(params)
        alert_filter = settings.get('AuditAlertFilter') or []

        exists = next((True for x in alert_filter if x['name'].lower() == name.lower()), False)
        if exists:
            raise CommandError('alert add', f'Alert name \"{name}\" is not unique')

        last_id = max((x['id'] for x in alert_filter), default=0)
        if not last_id:
            last_id = 0
        alert_id = last_id + 1

        alert = {
            'id': alert_id,
            'alertUid': secrets.randbelow(2**31),
            'name': name,
            'frequency': {
                'period': 'event'
            },
            'filter': {}
        }
        self.apply_alert_options(params, alert, **kwargs)

        rq = {
            'command': 'put_enterprise_setting',
            'type': 'AuditAlertFilter',
            'settings': alert,
        }
        api.communicate(params, rq)

        active = kwargs.get('active')
        if isinstance(active, str):
            if active == 'off':
                rq = {
                    'command': 'put_enterprise_setting',
                    'type': 'AuditAlertContext',
                    'settings': {
                        'id': alert_id,
                        'disabled': True
                    }
                }
                api.communicate(params, rq)

        self.invalidate_alerts()
        command = AuditAlertView()
        command.execute(params, target=str(alert_id))


class AuditAlertEdit(EnterpriseCommand, AuditSettingMixin):
    def get_parser(self):
        return alert_edit_parser

    def execute(self, params, **kwargs):
        alert = AuditSettingMixin.get_alert_configuration(params, kwargs.get('target'))
        self.apply_alert_options(params, alert, **kwargs)

        rq = {
            'command': 'put_enterprise_setting',
            'type': 'AuditAlertFilter',
            'settings': alert,
        }
        api.communicate(params, rq)

        active = kwargs.get('active')
        if isinstance(active, str):
            alert_id = alert.get('id')
            ctx = AuditSettingMixin.get_alert_context(alert_id) or {'id': alert_id}
            current_active = 'off' if ctx.get('disabled') is True else 'on'
            if active != current_active:
                rq = {
                    'command': 'put_enterprise_setting',
                    'type': 'AuditAlertContext',
                    'settings': {
                        'id': alert_id,
                        'disabled': active == 'off'
                    }
                }
                api.communicate(params, rq)

        self.invalidate_alerts()
        command = AuditAlertView()
        command.execute(params, target=kwargs.get('target'))


class AuditAlerts(GroupCommand):
    def __init__(self):
        super(AuditAlerts, self).__init__()
        self.register_command('list', AuditAlertList(), 'Display alert list', 'l')
        self.register_command('view', AuditAlertView(), 'View alert configuration', 'v')
        self.register_command('history', AuditAlertHistory(), 'View alert history', 'h')
        self.register_command('delete', AuditAlertDelete(), 'Delete audit alert(s) - single, range, or all', 'd')
        self.register_command('add', AuditAlertAdd(), 'Add audit alert', 'a')
        self.register_command('edit', AuditAlertEdit(), 'Edit audit alert', 'e')
        self.register_command('reset-counts', AuditAlertResetCount(), 'Reset alert counts')
        self.register_command('enable', AuditAlertSent('enable'), 'Enable audit alert')
        self.register_command('disable', AuditAlertSent('disable'), 'Disable audit alert')
        self.register_command('recipient', AuditAlertRecipients(), 'Modify alert recipients', 'r')
        self.default_verb = 'list'



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import io
import sys
import re
from pathlib import Path
from typing import Optional
from ..decorators.logging import logger, debug_decorator
from ...params import KeeperParams

class CommandHandler:
    @debug_decorator
    def execute_cli_command(self, params: KeeperParams, command: str) -> str:
        """Execute a CLI command and capture output."""
        output = io.StringIO()
        sys.stdout = output
        try:
            # Set service mode flag to bypass master password enforcement
            original_service_mode = getattr(params, 'service_mode', False)
            params.service_mode = True
            
            from ... import cli
            cli.do_command(params, command)
            return output.getvalue()
        except Exception as e:
            logger.debug(f"Error executing CLI command '{command}': {e}")
            return ''
        finally:
            sys.stdout = sys.__stdout__
            # Restore original service mode flag
            params.service_mode = original_service_mode

    @debug_decorator
    def find_config_record(self, params: KeeperParams, title: str) -> Optional[str]:
        """Find existing config record and return its UID."""
        try:
            output = self.execute_cli_command(params, f"search -v '{title}'")
            if uid_match := re.search(r'UID: ([a-zA-Z0-9_-]+)', output):
                return uid_match.group(1)
        except Exception as e:
            logger.error(f"Error searching for record: {e}")
            print(f"Error searching for record: {e}")
        return None

    @debug_decorator
    def get_help_output(self, params: KeeperParams) -> str:
        """Get help output from CLI."""
        output = io.StringIO()
        sys.stdout = output
        try:
            # Set service mode flag to bypass master password enforcement
            original_service_mode = getattr(params, 'service_mode', False)
            params.service_mode = True
            
            from ... import cli
            cli.do_command(params, 'help')
            return output.getvalue()
        finally:
            sys.stdout = sys.__stdout__
            # Restore original service mode flag
            params.service_mode = original_service_mode
    
    @debug_decorator
    def download_config_from_vault(self, params: KeeperParams, title: str, config_dir: Path) -> bool:
        """Download config file from vault if it exists."""
        try:
            if record_uid := self.find_config_record(params, title):
                self.execute_cli_command(params, f"download-attachment {record_uid} --out-dir '{config_dir}'")
                return True
            return False
        except Exception as e:
            logger.error(f"Error downloading config from vault: {e}")
            return False


#!/usr/bin/env python3

# Dumps HashiCorp Vault secrets to a Keeper JSON file ready for import.

# Requirements:
# pip install hvac

# Reads env vars:
#  VAULT_ADDR  which points to desired Hashicorp Vault instance, default http://localhost:8200
#  VAULT_TOKEN vault access token, no default - if empty get token from already-authenticated session
#  VAULT_MOUNT_PATH  to specify storage engine mount point, default /secret/
#  VAULT_ROOT_PREFIX to specify path to dump for partial backups, default '/'

# To use custom encoding set:
#  PYTHONIOENCODING=utf-8

# Example usage
# set PYTHONIOENCODING=utf-8
# set VAULT_ADDR=http://127.0.0.1:8200
# set VAULT_MOUNT_PATH=/secret
# set VAULT_ROOT_PREFIX=/dev/
# vault_dump_hvac.py > export.json

import copy
import datetime
import getpass
import json
import os
import subprocess
import sys

import hvac

def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)

def print_header():
    user = getpass.getuser()
    date = "{} UTC".format(datetime.datetime.utcnow())
    vault_address = os.environ.get('VAULT_ADDR', 'http://localhost:8200')
    vault_token = os.environ.get('VAULT_TOKEN','')
    vault_mount_path = os.environ.get('VAULT_MOUNT_PATH','/secret/')
    vault_root_prefix = os.environ.get('VAULT_ROOT_PREFIX','/')

    eprint ('# vault_dump_hvac.py')
    eprint ("# dump made by {}".format(user))
    eprint ("# backup date: {}".format(date))
    eprint ("# VAULT_ADDR env variable: {}".format(vault_address))
    eprint ("# VAULT_TOKEN env variable: {}".format('*' * len(vault_token)))
    eprint ("# VAULT_MOUNT_PATH env variable: {}".format(vault_mount_path))
    eprint ("# VAULT_ROOT_PREFIX env variable: {}".format(vault_root_prefix))
    eprint ('# STDIN encoding: {}'.format(sys.stdin.encoding))
    eprint ('# STDOUT encoding: {}'.format(sys.stdout.encoding))
    eprint ('#')
    eprint ('# WARNING: not guaranteed to be consistent!')
    eprint ()

# looks at an argument for a value and adds it to the export dict
def recurse_for_values(path_prefix, candidate_key, data_dict):
    # NB! if latest version is deleted - record is considered deleted by vault
    # although prev version(s) may still exist but not restored
    if 'keys' in candidate_key:
        candidate_values = candidate_key['keys']
    else:
        candidate_values = candidate_key['data']['keys']

    for candidate_value in candidate_values:
        next_index = path_prefix + candidate_value
        if candidate_value.endswith('/'):
            next_value = client.secrets.kv.v2.list_secrets(path=next_index)
            recurse_for_values(next_index, next_value, data_dict)
        else:
            deletion_time = ''
            final_data = {}

            try:
                final_data = client.secrets.kv.v2.read_secret(next_index) or {}
            except hvac.exceptions.InvalidPath as e:
                if e.args and isinstance(e.args, tuple) and len(e.args) > 0:
                    dict = json.loads(e.args[0])
                    deletion_time = dict.get('data', {}).get('metadata', {}).get('deletion_time', '')

            if 'data' in final_data:
                dirs = [s for s in next_index.split('/') if s != '']
                rec = copy.deepcopy(record_dict_template)
                rec['uid'] = len(data_dict['records']) + 1
                rec['title'] = dirs[-1]
                folder = '\\'.join(dirs[:-1])
                rec['folders'].append({'folder': folder})

                final_dict = final_data['data']
                final_data = final_dict.get('data') or {}
                # sorted_final_keys = sorted(final_data.keys())
                for final_key in final_data.keys():
                    field_key = field_key_template.format(final_key)
                    final_value = final_data[final_key]
                    field_value = final_value if isinstance(final_value, str) else json.dumps(final_value)
                    rec['custom_fields'][field_key] = field_value

                data_dict['records'].append(rec)
            elif 'rules' in final_data:
                rules = "\necho -ne {} | vault policy write {} -".format(repr(final_data['rules']), candidate_value)
                eprint("\n*** WARNING: skipped rules: " + rules)
            elif deletion_time:
                eprint("\n*** WARNING: index {} deleted {}".format(repr(next_index), repr(deletion_time)))
            else:
                eprint("\n*** WARNING: no data for {}".format(repr(next_index)))

env_vars = os.environ.copy()
hvac_token = os.environ.get('VAULT_TOKEN', '')
if not hvac_token:
    # Get token from already-authenticated session - needs vault binary in $PATH.
    eprint("\n*** WARNING: VAULT_TOKEN not present or empty")
    eprint("\n*** WARNING: Trying to get token from already-authenticated session - needs vault binary in $PATH.")
    hvac_token = subprocess.check_output("vault read -field id auth/token/lookup-self", shell=True, env=env_vars)

hvac_url = os.environ.get('VAULT_ADDR', 'http://localhost:8200')
client = hvac.Client(url=hvac_url, token=hvac_token)

if os.environ.get('VAULT_SKIP_VERIFY'):
    import requests
    rs = requests.Session()
    client.session = rs
    rs.verify = False
    import warnings
    warnings.filterwarnings("ignore")
assert client.is_authenticated()

vault_mount_path = os.environ.get('VAULT_MOUNT_PATH','/secret/').strip()
vault_root_prefix = os.environ.get('VAULT_ROOT_PREFIX','/').strip()
if not vault_root_prefix.startswith('/'):
        vault_root_prefix = '/' + vault_root_prefix
if not vault_root_prefix.endswith('/'):
    vault_root_prefix += '/'

field_key_template = "$text:{}:1"
record_dict_template = {
    "uid": None,
    "title": "",
    "$type": "login",
    "custom_fields": {
        # "$text:KEY1:1": "VALUE1",
        # "$text:KEY2:1": "VALUE2"
    },
    "folders": [
        # {"folder": "A\\B\\C\\D"}
    ]
}

print_header()

# must have right to list keys for VAULT_ROOT_PREFIX, otherwise get error with trace.
# NB! path must exist, ex. if vault is empty - error: InvalidPath
top_level_keys = client.secrets.kv.v2.list_secrets(mount_point=vault_mount_path, path=vault_root_prefix)
export_dict = {'records': []}
recurse_for_values(vault_root_prefix, top_level_keys, export_dict)

print(json.dumps(export_dict, indent=2))
print()



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

from ..service.commands.create_service import CreateService
from ..service.commands.config_operation import AddConfigService
from ..service.commands.handle_service import StartService, StopService, ServiceStatus

def register_commands(commands):
    commands['service-create'] = CreateService()
    commands['service-config-add'] = AddConfigService()
    commands['service-start'] = StartService()
    commands['service-stop'] = StopService()
    commands['service-status'] = ServiceStatus()

def register_command_info(aliases, command_info):
    service_classes = [
        CreateService,
        AddConfigService,
        StartService,
        StopService,
        ServiceStatus
    ]
    
    for service_class in service_classes:
        parser = service_class()
        p = parser.get_parser()
        command_info[p.prog] = p.description


import argparse
import logging
import sys
from typing import Optional, List

from keepercommander import api, utils, crypto
from keepercommander.__main__ import get_params_from_config
from keepercommander.proto import folder_pb2

parser = argparse.ArgumentParser(description='Add user to shared folder')
parser.add_argument('--config', dest='config', action='store', help='Config file to use')
parser.add_argument('--debug', dest='debug', action='store_true', help='Enables debug logging')
parser.add_argument('--teams', dest='teams', action='store_true', help='Load teams')
parser.add_argument('-p', '--manage-records', dest='manage_records', action='store', choices=['on', 'off'],
                    help='account permission: can manage records.')
parser.add_argument('-o', '--manage-users', dest='manage_users', action='store',  choices=['on', 'off'],
                    help='account permission: can manage users.')
parser.add_argument('--add', dest='add', action='append', help='Users to add to a shared folder.')
parser.add_argument('--remove', dest='remove', action='append', help='Users to remove from shared folder.')
parser.add_argument('shared_folder', help='Shared Folder UID')
opts, flags = parser.parse_known_args(sys.argv[1:])

logging.basicConfig(level=logging.DEBUG if opts.debug is True else logging.WARNING, format='%(message)s')

my_params = get_params_from_config(opts.config or '')

api.login(my_params)
if not my_params.session_token:
    exit(1)

shared_folder_uid = opts.shared_folder
team_uids = []   # type: List[str]
shared_folder_key = None   # type: Optional[bytes]
can_manage_users = False

if opts.teams is True:
    sd_rq = {
        'command': 'sync_down',
        'includes': ['teams', 'explicit']
    }
    sd_rs = api.communicate(my_params, sd_rq)
    for team in sd_rs.get('teams', []):
        if 'shared_folder_keys' in team:
            sf_key = next((x for x in team['shared_folder_keys'] if x['shared_folder_uid'] == shared_folder_uid), None)
            if sf_key:
                team_uids.append(team['team_uid'])
                if not shared_folder_key:
                    encrypted_key = utils.base64_url_decode(team['team_key'])
                    team_key_type = team['team_key_type']
                    try:
                        if team_key_type == 1:
                            team_key = crypto.decrypt_aes_v1(encrypted_key, my_params.data_key)
                        elif team_key_type == 2:
                            team_key = crypto.decrypt_rsa(encrypted_key, my_params.rsa_key2)
                        elif team_key_type == 3:
                            team_key = crypto.decrypt_aes_v2(encrypted_key, my_params.data_key)
                        elif team_key_type == 4:
                            team_key = crypto.decrypt_ec(encrypted_key, my_params.ecc_key)
                        else:
                            continue
                    except:
                        continue
                    encrypted_key = utils.base64_url_decode(sf_key['shared_folder_key'])
                    key_type = sf_key['key_type']
                    try:
                        if key_type == 1:
                            shared_folder_key = crypto.decrypt_aes_v1(encrypted_key, team_key)
                        elif key_type == 2:
                            if 'team_private_key' in team:
                                encrypted_rsa_key = utils.base64_url_decode(team['team_private_key'])
                                rsa_key = crypto.load_rsa_private_key(crypto.decrypt_aes_v1(encrypted_rsa_key, team_key))
                                shared_folder_key = crypto.decrypt_rsa(encrypted_key, rsa_key)
                        elif key_type == 3:
                            shared_folder_key = crypto.decrypt_aes_v2(encrypted_key, team_key)
                        elif key_type == 4:
                            if 'team_ecc_private_key' in team:
                                encrypted_ec_key = utils.base64_url_decode(team['team_ecc_private_key'])
                                ec_key = crypto.load_ec_private_key(crypto.decrypt_aes_v2(encrypted_ec_key, team_key))
                                shared_folder_key = crypto.decrypt_ec(encrypted_key, ec_key)
                    except:
                        continue

sf_rq = {
    'shared_folder_uid': shared_folder_uid
}
if len(team_uids) > 0:
    sf_rq['team_uid'] = team_uids[0]
rq = {
    'command': 'get_shared_folders',
    'shared_folders': [sf_rq],
    'include': ['sfheaders', 'sfusers', 'sfteams']
}
sf_rs = api.communicate(my_params, rq)
if len(sf_rs['shared_folders']) == 0:
    raise ValueError(f'Shared folder UID "{shared_folder_uid}" not found')

shared_folder_info = sf_rs['shared_folders'][0]
if not shared_folder_key:
    if 'shared_folder_key' in shared_folder_info:
        encrypted_key = utils.base64_url_decode(shared_folder_info['shared_folder_key'])
        key_type = shared_folder_info["key_type"]
        if key_type == 1:
            shared_folder_key = crypto.decrypt_aes_v1(encrypted_key, my_params.data_key)
        elif key_type == 2:
            shared_folder_key = crypto.decrypt_rsa(encrypted_key, my_params.rsa_key2)
        elif key_type == 3:
            shared_folder_key = crypto.decrypt_aes_v2(encrypted_key, my_params.data_key)
        elif key_type == 4:
            shared_folder_key = crypto.decrypt_ec(encrypted_key, my_params.ecc_key)

if not shared_folder_key:
    raise ValueError(f'Key cannot be resolved: shared folder UID "{shared_folder_uid}"')

team_uid = None   # type: Optional[str]
permissions = next((x for x in shared_folder_info.get('users', []) if x['username'] == my_params.user and x.get('manage_users') is True), None)
if permissions and len(team_uids) > 0:
    permissions = next((x for x in shared_folder_info.get('teams', []) if x['team_uid'] in team_uids and x.get('manage_users') is True), None)
    if permissions:
        team_uid = permissions['team_uid']

existing_users = set()
if isinstance(shared_folder_info.get('users'), list):
    existing_users.update((x.get('username').lower() for x in shared_folder_info['users']))

users_to_add = []
if hasattr(opts, 'add'):
    if isinstance(opts.add, list):
        for email in opts.add:
            email = email.lower()
            if email in existing_users:
                logging.warning('Add user "%s": already belongs to the shared folder', email)
            else:
                users_to_add.append(email)

users_to_remove = []
if hasattr(opts, 'remove'):
    if isinstance(opts.remove, list):
        for email in opts.remove:
            email = email.lower()
            if email in existing_users:
                users_to_remove.append(email)
            else:
                logging.warning('Remove user "%s": does not belongs to the shared folder', email)

if len(users_to_add) > 0:
    public_keys = {x: None for x in users_to_add}
    api.load_user_public_keys(my_params, users_to_add, False)

manage_users = None
manage_records = None
if opts.manage_users:
    manage_users = opts.manage_users == 'on'
if opts.manage_records:
    manage_records = opts.manage_records == 'on'

rq = folder_pb2.SharedFolderUpdateV3Request()
rq.sharedFolderUid = utils.base64_url_decode(opts.shared_folder)
rq.forceUpdate = True

for user in users_to_add:
    arq = folder_pb2.SharedFolderUpdateUser()
    arq.username = user
    if isinstance(manage_users, bool):
        arq.manageUsers = folder_pb2.BOOLEAN_TRUE if manage_users else folder_pb2.BOOLEAN_FALSE
    if isinstance(manage_records, bool):
        arq.manageRecords = folder_pb2.BOOLEAN_TRUE if manage_records else folder_pb2.BOOLEAN_FALSE
    public_keys = my_params.key_cache.get(user)
    if public_keys:
        if public_keys.ec:
            user_ec_key = crypto.load_ec_public_key(public_keys.ec)
            arq.typedSharedFolderKey.encryptedKey = crypto.encrypt_ec(shared_folder_key, user_ec_key)
            arq.typedSharedFolderKey.encryptedKeyType = folder_pb2.encrypted_by_public_key_ecc
            rq.sharedFolderAddUser.append(arq)
        elif not my_params.forbid_rsa and public_keys.rsa:
            user_rsa_key = crypto.load_rsa_public_key(public_keys.rsa)
            arq.typedSharedFolderKey.encryptedKey = crypto.encrypt_rsa(shared_folder_key, user_rsa_key)
            arq.typedSharedFolderKey.encryptedKeyType = folder_pb2.encrypted_by_public_key
            rq.sharedFolderAddUser.append(arq)
    else:
        logging.warning('Add user "%s": User public key is not available', user)

for user in users_to_remove:
    rq.sharedFolderRemoveUser.append(user)

rs = api.communicate_rest(my_params, rq, 'vault/shared_folder_update_v3', rs_type=folder_pb2.SharedFolderUpdateV3Response)
for add_status in rs.sharedFolderAddUserStatus:
    if add_status.status != 'success':
        logging.warning(f'Failed to add user {add_status.username} to shared folder: {add_status.status}')
for remove_status in rs.sharedFolderRemoveUserStatus:
    if remove_status.status != 'success':
        logging.warning(f'Failed to add user {remove_status.username} to shared folder: {remove_status.status}')



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

from typing import Optional
from ...params import KeeperParams
from ... import utils

_current_params: Optional[KeeperParams] = None

def init_globals(params: KeeperParams) -> None:
    global _current_params
    _current_params = params

def get_current_params() -> Optional[KeeperParams]:
    return _current_params

def ensure_params_loaded() -> KeeperParams:
    """Load params from config if not already loaded."""
    params = get_current_params()
    if not params: 
        from ...__main__ import get_params_from_config
        config_path = utils.get_default_path() / "config.json"
        params = get_params_from_config(config_path)
        init_globals(params)
    return params


#!/usr/bin/env python3
"""
Convert LastPass shared folder membership format to Keeper format.

This script converts shared folder data from LastPass (getsfdata format)
to Keeper's shared folder format with separate teams section.
"""

import argparse
import json
import os
import sys
from collections import defaultdict
from typing import Dict, Any


def convert_folder_name(lastpass_name: str) -> str:
    """
    Convert LastPass folder name to Keeper format.
    Keeper uses "Shared-" prefix format.
    """
    if not lastpass_name.startswith("Shared-"):
        return f"Shared-{lastpass_name}"
    return lastpass_name


def convert_lastpass_to_keeper(lastpass_data: Dict[str, Any]) -> Dict[str, Any]:
    """
    Convert LastPass shared folder data to Keeper format.

    Args:
        lastpass_data: Dictionary with folder IDs as keys and folder data as values

    Returns:
        Dictionary in Keeper format with teams and shared_folders sections
    """
    # Track teams and their members
    teams = defaultdict(set)  # team_name -> set of member emails
    shared_folders = []

    for folder_id, folder_data in lastpass_data.items():
        # Skip deleted folders
        if folder_data.get("deleted", False):
            continue

        folder_name = folder_data.get("sharedfoldername", f"Folder-{folder_id}")
        keeper_folder_name = convert_folder_name(folder_name)

        # Track permissions and groups for this folder
        permissions = []
        folder_groups = set()  # Groups that have permissions in this folder

        users = folder_data.get("users", [])
        for entry in users:
            # Skip deleted entries
            if entry.get("deletedstatus", "0") == "1":
                continue

            username = entry.get("username")
            if not username:
                print(f"Warning: Skipping entry with no username in folder {folder_name}", file=sys.stderr)
                continue

            readonly = entry.get("readonly", "1") == "1"
            can_administer = entry.get("can_administer", "0") == "1"
            manage_records = not readonly or can_administer

            if "group_name" in entry:
                # This is a group permission entry
                group_name = entry["group_name"]

                # Add user to the team
                teams[group_name].add(username)

                # Add group permission only once per folder
                if group_name not in folder_groups:
                    folder_groups.add(group_name)
                    permissions.append({
                        "name": group_name,
                        "manage_users": can_administer,
                        "manage_records": manage_records
                    })
            else:
                # This is a direct user permission entry
                permissions.append({
                    "name": username,
                    "manage_users": can_administer,
                    "manage_records": manage_records
                })

        # Only add folder if it has active permissions
        if permissions:
            shared_folders.append({
                "path": keeper_folder_name,
                "permissions": permissions
            })

    # Convert teams defaultdict to list format
    teams_list = []
    for team_name, members in teams.items():
        teams_list.append({
            "name": team_name,
            "members": sorted(list(members))  # Sort for consistent output
        })

    # Sort teams by name for consistent output
    teams_list.sort(key=lambda x: x["name"])

    result = {
        "shared_folders": shared_folders
    }

    # Only add teams section if there are teams
    if teams_list:
        result["teams"] = teams_list

    return result


def main():
    parser = argparse.ArgumentParser(
        description="Convert LastPass shared folder format to Keeper format with teams"
    )
    parser.add_argument(
        "input_file",
        help="Input LastPass JSON file (getsfdata format)"
    )
    parser.add_argument(
        "--pretty",
        action="store_true",
        help="Pretty print the output JSON"
    )
    parser.add_argument(
        "--force",
        action="store_true",
        help="Overwrite output file if it exists"
    )

    args = parser.parse_args()

    # Fixed output filename
    output_file = "shared_folder_membership.json"

    try:
        # Check if output file exists
        if os.path.exists(output_file) and not args.force:
            print(f"Error: Output file '{output_file}' already exists!", file=sys.stderr)
            print(f"Use --force to overwrite the existing file", file=sys.stderr)
            sys.exit(1)

        # Read input file
        with open(args.input_file, 'r', encoding='utf-8') as f:
            lastpass_data = json.load(f)

        # Validate input format
        if not isinstance(lastpass_data, dict):
            print("Error: Input file should contain a JSON object with folder IDs as keys", file=sys.stderr)
            sys.exit(1)

        # Convert to Keeper format
        keeper_data = convert_lastpass_to_keeper(lastpass_data)

        # Write output file
        with open(output_file, 'w', encoding='utf-8') as f:
            if args.pretty:
                json.dump(keeper_data, f, indent=2, ensure_ascii=False)
            else:
                json.dump(keeper_data, f, ensure_ascii=False)

        # Print summary
        folder_count = len(keeper_data["shared_folders"])
        team_count = len(keeper_data.get("teams", []))
        total_permissions = sum(len(folder["permissions"]) for folder in keeper_data["shared_folders"])
        total_team_members = sum(len(team["members"]) for team in keeper_data.get("teams", []))

        print(f"Conversion completed successfully!")
        print(f"Converted {folder_count} shared folders with {total_permissions} total permissions")
        print(f"Created {team_count} teams with {total_team_members} total team memberships")
        print(f"Output written to: {output_file}")

    except FileNotFoundError:
        print(f"Error: Input file '{args.input_file}' not found", file=sys.stderr)
        sys.exit(1)
    except json.JSONDecodeError as e:
        print(f"Error: Invalid JSON in input file: {e}", file=sys.stderr)
        sys.exit(1)
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        sys.exit(1)


def print_conversion_info():
    """Print information about the conversion process."""
    print("LastPass to Keeper Shared Folder Converter")
    print("=" * 50)
    print()
    print("This script converts LastPass shared folder data to Keeper format.")
    print()
    print("Key Features:")
    print("- Extracts teams from group permissions and creates separate teams section")
    print("- Handles both direct user permissions and group-based permissions")
    print("- Each team appears only once per shared folder")
    print("- Teams are automatically populated with members based on group permissions")
    print()
    print("LastPass Input Structure:")
    print("- Folder entries contain 'users' array")
    print("- Group permissions have both 'username' and 'group_name' fields")
    print("- Direct user permissions have only 'username' field")
    print()
    print("Keeper Output Structure:")
    print("- 'teams' section with team names and member lists")
    print("- 'shared_folders' section with permissions referencing teams by name")
    print()
    print("Permission Mapping:")
    print("- LastPass 'readonly': '0' -> Keeper 'manage_records': true")
    print("- LastPass 'readonly': '1' -> Keeper 'manage_records': false")
    print("- LastPass 'can_administer': '1' -> Keeper 'manage_users': true")
    print("- Admins get manage_records=true regardless of readonly status")
    print()
    print("Usage:")
    print("  python converter.py input.json [--pretty] [--force]")
    print()
    print("Options:")
    print("  --pretty    Pretty print the JSON output")
    print("  --force     Overwrite shared_folder_membership.json if it exists")
    print()
    print("Output: Always creates 'shared_folder_membership.json'")


if __name__ == "__main__":
    if len(sys.argv) == 1:
        print_conversion_info()
    else:
        main()


#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

from functools import wraps
from typing import Callable, Any
from flask import request
import time
import re
from .logging import logger

def sanitize_password_in_command(data):
    """Sanitize password values in command string and filedata"""
    if not data:
        return data
    
    sanitized = data.copy()
    
    # Sanitize command string if present
    if 'command' in sanitized:
        command = sanitized['command']
        # Pattern to match password=value (with or without quotes)
        password_pattern = r"password=(['\"]?)([^'\"\s]{1,1024})\1"
        sanitized['command'] = re.sub(password_pattern, r"password=\1***\1", command)
    
    # Sanitize filedata if present
    if 'filedata' in sanitized:
        sanitized['filedata'] = _sanitize_nested_data(sanitized['filedata'])
    
    return sanitized

def _sanitize_nested_data(data):
    """Recursively sanitize nested data structures"""
    if isinstance(data, dict):
        sanitized = {}
        for key, value in data.items():
            # Sanitize sensitive field names
            if key.lower() in ['password', 'login', 'secret', 'token', 'key']:
                if isinstance(value, str) and len(value) > 0:
                    sanitized[key] = '*' * min(len(value), 15)
                else:
                    sanitized[key] = '***'
            else:
                sanitized[key] = _sanitize_nested_data(value)
        return sanitized
    elif isinstance(data, list):
        return [_sanitize_nested_data(item) for item in data]
    elif isinstance(data, str):
        # Sanitize email addresses in string values to protect PII
        import re
        sanitized_str = re.sub(r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b', '***@***.***', data)
        return sanitized_str
    else:
        return data

def _get_sanitized_request_data():
    """Extract and sanitize request data for logging (only for JSON POST requests)"""
    sanitized_data = None
    if request.method == 'POST' and request.content_type and 'application/json' in request.content_type:
        try:
            json_data = request.get_json(silent=True)
            sanitized_data = sanitize_password_in_command(json_data)
        except Exception:
            sanitized_data = None
    return f"data={sanitized_data}" if sanitized_data else "no-data"

def api_log_handler(fn: Callable) -> Callable:
    """Log API request information in a single line with color-coded status"""
    @wraps(fn)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        ip = request.headers.get('X-Forwarded-For', request.remote_addr)
        if ip is None:
            ip = request.remote_addr
            
        start_time = time.time()
        try:
            result = fn(*args, **kwargs)
            duration = time.time() - start_time
            
            status_code = result[1] if isinstance(result, tuple) else 200
            
            # Color code based on status
            if status_code < 200:
                status_color = "\033[94m"  # Blue for informational
            elif status_code < 300:
                status_color = "\033[92m"  # Green for success
            elif status_code < 400:
                status_color = "\033[96m"  # Cyan for redirection
            elif status_code < 500:
                status_color = "\033[91m"  # Red for client errors
            else:
                status_color = "\033[93m"  # Yellow for server errors
            
            # Sanitize request data to hide passwords (only for JSON requests)
            data_str = _get_sanitized_request_data()
            
            log_parts = [
                f"\033[94m{request.method}\033[0m",  # Blue method
                f"\033[96m{request.path}\033[0m",    # Cyan path
                f"\033[95m{ip}\033[0m",              # Magenta IP
                data_str,
                str(status_code),
                f"{duration:.2f}s"
            ]
            
            logger.info(" | ".join(log_parts))
            return result
            
        except Exception as ex:
            duration = time.time() - start_time
            
            # Sanitize request data for error logs too (only for JSON requests)
            data_str = _get_sanitized_request_data()
            
            log_parts = [
                request.method,
                request.path,
                ip,
                data_str,
                "500",
                f"{duration:.2f}s",
                f"error='{str(ex)}'"
            ]
            logger.error(" | ".join(log_parts))
            raise
            
    return wrapper


from __future__ import annotations
import logging
from .constants import DIS_INFRA_GRAPH_ID
from .utils import get_connection, make_agent
from ..keeper_dag import DAG, EdgeType
from ..keeper_dag.exceptions import DAGVertexException
from ..keeper_dag.crypto import urlsafe_str_to_bytes
import os
import importlib
import time
from typing import Any, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from ..keeper_dag.vertex import DAGVertex


class Infrastructure:

    """
    Create a graph of the infrastructure.

    The first run will create a full graph since the vertices do not exist.
    Further discovery run will only show vertices that ...
    * do not have vaults records.
    * the data has changed.
    * the ACL has changed.

    """

    KEY_PATH = "infrastructure"
    DELTA_PATH = "delta"
    ADMIN_PATH = "ADMINS"
    USER_PATH = "USERS"

    def __init__(self, record: Any, logger: Optional[Any] = None, history_level: int = 0,
                 debug_level: int = 0, fail_on_corrupt: bool = True, log_prefix: str = "GS Infrastructure",
                 save_batch_count: int = 200, agent: Optional[str] = None,
                 **kwargs):

        # This will either be a KSM Record, or Commander KeeperRecord
        self.record = record
        self._dag = None
        if logger is None:
            logger = logging.getLogger()
        self.logger = logger
        self.log_prefix = log_prefix
        self.history_level = history_level
        self.debug_level = debug_level
        self.fail_on_corrupt = fail_on_corrupt
        self.save_batch_count = save_batch_count

        self.auto_save = False
        self.delta_graph = True
        self.last_sync_point = -1

        self.agent = make_agent("infra")
        if agent is not None:
            self.agent += "; " + agent

        self.conn = get_connection(**kwargs)

    @property
    def dag(self) -> DAG:
        if self._dag is None:

            self.logger.debug(f"loading the dag graph {DIS_INFRA_GRAPH_ID}")
            self.logger.debug(f"setting graph save batch count to {self.save_batch_count}")

            self._dag = DAG(conn=self.conn, record=self.record, graph_id=DIS_INFRA_GRAPH_ID, auto_save=self.auto_save,
                            logger=self.logger, history_level=self.history_level, debug_level=self.debug_level,
                            name="Discovery Infrastructure", fail_on_corrupt=self.fail_on_corrupt,
                            log_prefix=self.log_prefix, save_batch_count=self.save_batch_count,
                            agent=self.agent)
            # Do not load the DAG here.
            # We don't know if we are using a sync point yet.

        return self._dag

    @property
    def has_discovery_data(self) -> bool:
        # Does the graph array have any vertices?
        if self.dag.has_graph is False:
            return False

        # If we at least have the root, does is have the configuration?
        if self.get_root.has_vertices() is False:
            return False

        return True

    @property
    def get_root(self) -> DAGVertex:
        return self.dag.get_root

    @property
    def get_configuration(self) -> DAGVertex:
        try:
            configuration = self.get_root.has_vertices()[0]
        except (Exception,):
            raise DAGVertexException("Could not find the configuration vertex for the infrastructure graph.")
        return configuration

    @property
    def sync_point(self):
        return self._dag.load(sync_point=0)

    def load(self, sync_point: int = 0):
        ts = time.time()
        res = self.dag.load(sync_point=sync_point) or 0
        self.logger.debug(f"infrastructure took {time.time()-ts} secs to load")
        return res

    def save(self, delta_graph: Optional[bool] = None):
        if delta_graph is None:
            delta_graph = self.delta_graph

        self.logger.debug(f"current sync point {self.last_sync_point}")
        if delta_graph is True:
            self.logger.debug("saving delta graph of the infrastructure")
        ts = time.time()
        self._dag.save(delta_graph=delta_graph)
        self.logger.debug(f"infrastructure took {time.time()-ts} secs to save")

    def to_dot(self, graph_format: str = "svg", show_hex_uid: bool = False,
               show_version: bool = True, show_only_active_vertices: bool = False,
               show_only_active_edges: bool = False, sync_point: int = None, graph_type: str = "dot"):

        try:
            mod = importlib.import_module("graphviz")
        except ImportError:
            raise Exception("Cannot to_dot(), graphviz module is not installed.")

        dot = getattr(mod, "Digraph")(comment=f"DAG for Discovery", format=graph_format)

        if sync_point is None:
            sync_point = self.last_sync_point

        self.logger.debug(f"generating infrastructure dot starting at sync point {sync_point}")

        self.dag.load(sync_point=sync_point)

        count = 0
        if len(self.dag.get_root.has_vertices()) > 0:
            config_vertex = self.dag.get_root.has_vertices()[0]
            count = len(config_vertex.has_vertices())

        if graph_type == "dot":
            dot.attr(rankdir='RL')
            rank_sep = 10
            if count > 10:
                rank_sep += int(count * 0.10)
            dot.attr(ranksep=str(rank_sep))
        elif graph_type == "twopi":
            rank_sep = 20
            if count > 20:
                rank_sep += int(count * 0.10)

            dot.attr(layout="twopi")
            dot.attr(ranksep=str(rank_sep))
            dot.attr(ratio="auto")
        else:
            dot.attr(layout=graph_type)
            dot.attr(ranksep=10)

        for v in self.dag.all_vertices:
            if show_only_active_vertices is True and v.active is False:
                continue

            shape = "ellipse"
            fillcolor = "white"
            color = "black"

            if v.corrupt is False:

                if v.active is False:
                    fillcolor = "grey"

                record_type = None
                record_uid = None
                name = v.name
                source = None
                try:
                    data = v.content_as_dict
                    record_type = data.get("record_type")
                    record_uid = data.get("record_uid")
                    name = data.get("name")
                    item = data.get("item")
                    if item is not None:
                        if item.get("managed", False) is True:
                            shape = "box"
                        source = item.get("source")
                    if record_uid is not None:
                        fillcolor = "#AFFFAF"
                    if data.get("ignore_object", False) is True:
                        fillcolor = "#DFDFFF"
                except (Exception,):
                    pass

                label = f"uid={v.uid}"
                if record_type is not None:
                    label += f"\\nrt={record_type}"
                if name is not None and name != v.uid:
                    name = name.replace("\\", "\\\\")
                    label += f"\\nname={name}"
                if source is not None:
                    label += f"\\nsource={source}"
                if record_uid is not None:
                    label += f"\\nruid={record_uid}"
                if show_hex_uid is True:
                    label += f"\\nhex={urlsafe_str_to_bytes(v.uid).hex()}"
                if v.uid == self.dag.get_root.uid:
                    fillcolor = "gold"
                    label += f"\\nsp={sync_point}"

                tooltip = f"ACTIVE={v.active}\\n\\n"
                try:
                    content = v.content_as_dict
                    for k, val in content.items():
                        if k == "item":
                            continue
                        if isinstance(val, str) is True:
                            val = val.replace("\\", "\\\\")
                        tooltip += f"{k}={val}\\n"

                    item = content.get("item")
                    if item is not None:
                        tooltip += f"------------------\\n"
                        for k, val in item.items():
                            if isinstance(val, str) is True:
                                val = val.replace("\\", "\\\\")
                            tooltip += f"{k}={val}\\n"
                except Exception as err:
                    tooltip += str(err)
            else:
                fillcolor = "red"
                label = f"{v.uid} (CORRUPT)"
                tooltip = "CORRUPT"

            dot.node(v.uid, label, color=color, fillcolor=fillcolor, style="filled", shape=shape, tooltip=tooltip)

            head_uids = []
            for edge in v.edges:

                # Don't show edges that reference self, DATA and data that has been DELETION
                if edge.head_uid == v.uid:
                    continue

                if edge.head_uid not in head_uids:
                    head_uids.append(edge.head_uid)

            def _render_edge(e):

                color = "grey"
                style = "solid"

                if e.corrupt is False:

                    # To reduce the number of edges, only show the active edges
                    if e.active is True:
                        color = "black"
                        style = "bold"
                    elif show_only_active_edges is True:
                        return

                    # If the vertex is not active, gray out the DATA edge
                    if e.edge_type == EdgeType.DATA and v.active is False:
                        color = "grey"

                    if e.edge_type == EdgeType.DELETION:
                        style = "dotted"

                    edgetip = ""
                    if e.edge_type == EdgeType.ACL and v.active is True:
                        content = e.content_as_dict
                        for k, val in content.items():
                            edgetip += f"{k}={val}\\n"
                        if content.get("is_admin") is True:
                            color = "red"

                    label = DAG.EDGE_LABEL.get(e.edge_type)
                    if label is None:
                        label = "UNK"
                    if e.path is not None and e.path != "":
                        label += f"\\npath={e.path}"
                    if show_version is True:
                        label += f"\\nv={e.version}"
                else:
                    label = f"{e.edge_type.value} (CORRUPT)"
                    color = "red"
                    edgetip = "CORRUPT"

                # tail, head (arrow side), label, ...
                dot.edge(v.uid, e.head_uid, label, style=style, fontcolor=color, color=color, tooltip=edgetip)

            for head_uid in head_uids:
                version, edge = v.get_highest_edge_version(head_uid)
                _render_edge(edge)

            data_edge = v.get_data()
            if data_edge is not None:
                _render_edge(data_edge)

        return dot

    def render(self, name: str, **kwargs):

        output_name = os.environ.get("GRAPH_DIR", os.environ.get("HOME", os.environ.get("PROFILENAME", ".")))
        output_name = os.path.join(output_name, name)
        dot = self.to_dot(**kwargs)
        dot.render(output_name)



from __future__ import annotations
from .constants import DIS_RULES_GRAPH_ID
from .types import (RuleTypeEnum, RuleItem, ActionRuleSet, ActionRuleItem, ScheduleRuleSet, ComplexityRuleSet,
                    Statement, RuleActionEnum)
from .utils import value_to_boolean, get_connection, make_agent
from ..keeper_dag import DAG, EdgeType
from ..keeper_dag.exceptions import DAGException
from time import time
import base64
import os
from typing import Any, List, Optional, Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from .types import DiscoveryObject


class Rules:

    DATA_PATH = "rules"
    RULE_ITEM_TYPE_MAP = {
        "ActionRuleItem": RuleTypeEnum.ACTION,
        "ScheduleRuleItem": RuleTypeEnum.SCHEDULE,
        "ComplexityRuleItem": RuleTypeEnum.COMPLEXITY
    }
    RULE_TYPE_TO_SET_MAP = {
        RuleTypeEnum.ACTION: ActionRuleSet,
        RuleTypeEnum.SCHEDULE: ScheduleRuleSet,
        RuleTypeEnum.COMPLEXITY: ComplexityRuleSet
    }

    RULE_FIELDS = {
        # Attributes the records
        "recordType": {"type": str},
        "parentRecordType": {"type": str},
        "recordTitle": {"type": str},
        "recordNotes": {"type": str},
        "recordDesc": {"type": str},
        "parentUid": {"type": str},

        # Record fields
        "login": {"type": str},
        "password": {"type": str},
        "privatePEMKey": {"type": str},
        "distinguishedName": {"type": str},
        "connectDatabase": {"type": str},
        "managed": {"type": bool, "default": False},
        "hostName": {"type": str},
        "port": {"type": float, "default": 0},
        "operatingSystem": {"type": str},
        "instanceName": {"type": str},
        "instanceId": {"type": str},
        "providerGroup": {"type": str},
        "providerRegion": {"type": str},
        "databaseId": {"type": str},
        "databaseType": {"type": str},
        "useSSL": {"type": bool, "default": False},
        "domainName": {"type": str},
        "directoryId": {"type": str},
        "directoryType": {"type": str},
    }

    BREAK_OUT = {
        "pamHostname": {
            "hostName": "hostName",
            "port": "port"
        }
    }

    RECORD_FIELD = {
        "pamMachine": ["pamHostname"],
        "pamDatabase": ["pamHostname", "databaseType"],
        "pamDirectory": ["pamHostname", "directoryType"],
        "pamUser": ["parentUid", "login", "distinguishedName"],
    }

    OBJ_ATTR = {
        "parentUid": "parent_record_uid"
    }

    def __init__(self, record: Any, logger: Optional[Any] = None,  debug_level: int = 0, fail_on_corrupt: bool = True,
                 agent: Optional[str] = None, **kwargs):

        self.conn = get_connection(**kwargs)

        # This will either be a KSM Record, or Commander KeeperRecord
        self.record = record
        self._dag = None
        self.logger = logger
        self.debug_level = debug_level
        self.fail_on_corrupt = fail_on_corrupt

        self.agent = make_agent("rules")
        if agent is not None:
            self.agent += "; " + agent

    @property
    def dag(self) -> DAG:
        if self._dag is None:

            # Turn auto_save on after the DAG has been created.
            # No need to call it six times in a row to initialize it.
            self._dag = DAG(conn=self.conn, record=self.record, graph_id=DIS_RULES_GRAPH_ID, auto_save=False,
                            logger=self.logger, debug_level=self.debug_level, fail_on_corrupt=self.fail_on_corrupt,
                            agent=self.agent)
            self._dag.load()

            # Has the status been initialized?
            if self._dag.has_graph is False:
                for rule_type_enum in Rules.RULE_TYPE_TO_SET_MAP:
                    rules = self._dag.add_vertex()
                    rules.belongs_to_root(
                        EdgeType.KEY,
                        path=rule_type_enum.value
                    )
                    content = Rules.RULE_TYPE_TO_SET_MAP[rule_type_enum]()
                    rules.add_data(
                        content=content,
                    )
                self._dag.save()

            # The graph exists now, turn on the auto_save.
            self._dag.auto_save = True
        return self._dag

    @staticmethod
    def data_path(rule_type: RuleTypeEnum):
        return f"/{rule_type.value}"

    def get_ruleset(self, rule_type: RuleTypeEnum):
        path = self.data_path(rule_type)
        rule_json = self.dag.walk_down_path(path).content_as_str
        if rule_json is None:
            raise DAGException("Could not get the status data from the DAG.")
        rule_set_class = Rules.RULE_TYPE_TO_SET_MAP[rule_type]
        return rule_set_class.model_validate_json(rule_json)

    def set_ruleset(self, rule_type: RuleTypeEnum, rules: List[Rules]):
        path = self.data_path(rule_type)
        self.dag.walk_down_path(path).add_data(
            content=rules,
        )
        # Auto save should save the data

    def _rule_transaction(self, func: Callable, rule: Optional[RuleItem] = None):
        rule_type = rule.__class__.__name__
        rule_type_enum = Rules.RULE_ITEM_TYPE_MAP.get(rule_type)
        if rule_type_enum is None:
            raise ValueError("rule is not a known rule instance")

        # Get the ruleset and the rule list for the type
        ruleset = self.get_ruleset(rule_type_enum)

        # Call the specialized code
        rules = func(
            r=rule,
            rs=ruleset.rules
        )

        # Sort the rule by priority in asc order.
        ruleset.rules = list(sorted(rules, key=lambda x: x.priority))
        self.set_ruleset(rule_type_enum, ruleset)

    def add_rule(self, rule: RuleItem) -> RuleItem:

        if rule.rule_id is None:
            rule.rule_id = "RULE" + base64.urlsafe_b64encode(os.urandom(8)).decode().rstrip('=')
        if rule.added_ts is None:
            rule.added_ts = int(time())

        def _add_rule(r: RuleItem, rs: List[RuleItem]):
            rs.append(r)
            return rs

        self._rule_transaction(
            rule=rule,
            func=_add_rule
        )

        return rule

    def update_rule(self, rule: RuleItem) -> RuleItem:

        def _update_rule(r: RuleItem, rs: List[RuleItem]):
            new_rule_list = []
            for _r in rs:
                if _r.rule_id == r.rule_id:
                    new_rule_list.append(r)
                else:
                    new_rule_list.append(_r)
            return new_rule_list

        self._rule_transaction(
            rule=rule,
            func=_update_rule
        )

        return rule

    def remove_rule(self, rule: RuleItem):

        def _remove_rule(r: RuleItem, rs: List[RuleItem]):
            new_rule_list = []
            for _r in rs:
                if _r.rule_id != r.rule_id:
                    new_rule_list.append(_r)
            return new_rule_list

        self._rule_transaction(
            rule=rule,
            func=_remove_rule
        )

    def rule_list(self, rule_type: RuleTypeEnum, search: Optional[str] = None) -> List[RuleItem]:
        rule_list = []
        for rule_item in self.get_ruleset(rule_type).rules:
            if search is not None and rule_item.search(search) is False:
                continue
            rule_list.append(rule_item)

        return rule_list

    def get_rule_item(self, rule_type: RuleTypeEnum, rule_id: str) -> Optional[RuleItem]:
        for rule_item in self.rule_list(rule_type=rule_type):
            if rule_item.rule_id == rule_id:
                return rule_item
        return None

    @staticmethod
    def make_action_rule_from_content(content: DiscoveryObject, action: RuleActionEnum, priority: Optional[int] = None,
                                      case_sensitive: bool = True,
                                      shared_folder_uid: Optional[str] = None) -> ActionRuleItem:

        if action == RuleActionEnum.IGNORE:
            priority = -1

        record_fields = Rules.RECORD_FIELD.get(content.record_type)
        if record_fields is None:
            raise ValueError(f"Record type {content.record_type} does not have fields maps.")

        statements = [
            Statement(field="recordType", operator="==", value=content.record_type)
        ]

        for field_label in record_fields:
            if field_label in Rules.OBJ_ATTR:
                attr = Rules.OBJ_ATTR[field_label]
                if hasattr(content, attr) is False:
                    raise Exception(f"Discovery object is missing attribute {attr}")
                value = getattr(content, attr)
                statements.append(
                    Statement(field=field_label, operator="==", value=value)
                )
            else:
                for field in content.fields:
                    label = field.label
                    if field_label != label:
                        continue

                    value = field.value
                    if value is None or len(value) == 0:
                        continue
                    value = value[0]

                    if label in Rules.BREAK_OUT:
                        for key in Rules.BREAK_OUT[label]:
                            key_value = value.get(key)
                            if key_value is None:
                                continue
                            statements.append(
                                Statement(field=key, operator="==", value=key_value)
                            )
                    else:
                        statements.append(
                            Statement(field=label, operator="==", value=value)
                        )

        return ActionRuleItem(
            enabeld=True,
            priority=priority,
            case_sensitive=case_sensitive,
            statement=statements,
            action=action,
            shared_folder_uid=shared_folder_uid
        )

    @staticmethod
    def make_action_rule_statement_str(statement: List[Statement]) -> str:
        statement_str = ""
        for item in statement:
            if statement_str != "":
                statement_str += " and "
            statement_str += item.field + " " + item.operator + " "
            field_type = Rules.RULE_FIELDS.get(item.field).get("type")
            if field_type is None:
                raise ValueError("Unknown field in rule")
            if field_type is str:
                statement_str += f"'{item.value}'"
            elif field_type is bool:
                if value_to_boolean(item.value) is True:
                    statement_str += "true"
                else:
                    statement_str += "false"
            elif field_type is float:
                if int(item.value) == item.value:
                    statement_str += str(int(item.value))
                else:
                    statement_str += str(item.value)
            else:
                raise ValueError("Cannot determine the field type for rule statement.")
        return statement_str



from __future__ import annotations
import argparse
from . import PAMGatewayActionDiscoverCommandBase, GatewayContext
from .rule_add import PAMGatewayActionDiscoverRuleAddCommand
from ..pam.router_helper import router_get_connected_gateways
from ...display import bcolors
from ...discovery_common.rule import Rules, RuleTypeEnum


class PAMGatewayActionDiscoverRuleUpdateCommand(PAMGatewayActionDiscoverCommandBase):
    parser = argparse.ArgumentParser(prog='pam-action-discover-rule-update')
    parser.add_argument('--gateway', '-g', required=True, dest='gateway', action='store',
                        help='Gateway name of UID.')
    parser.add_argument('--rule-id', '-i', required=True, dest='rule_id', action='store',
                        help='Identifier for the rule')
    parser.add_argument('--action', '-a', required=False, choices=['add', 'ignore', 'prompt'],
                        dest='rule_action', action='store', help='Update the action to take if rule matches')
    parser.add_argument('--priority', '-p', required=False, dest='priority', action='store', type=int,
                        help='Update the rule execute priority')
    parser.add_argument('--ignore-case', required=False, dest='ignore_case', action='store_true',
                        help='Update the rule to ignore case')
    parser.add_argument('--no-ignore-case', required=False, dest='ignore_case', action='store_false',
                        help='Update the rule to not ignore case')
    parser.add_argument('--shared-folder-uid', required=False, dest='shared_folder_uid',
                        action='store', help='Update the folder to place record.')
    parser.add_argument('--statement', '-s', required=False, dest='statement', action='store',
                        help='Update the rule statement')

    def get_parser(self):
        return PAMGatewayActionDiscoverRuleUpdateCommand.parser

    def execute(self, params, **kwargs):

        if not hasattr(params, 'pam_controllers'):
            router_get_connected_gateways(params)

        gateway = kwargs.get("gateway")
        gateway_context = GatewayContext.from_gateway(params, gateway)
        if gateway_context is None:
            print(f'{bcolors.FAIL}Discovery job gateway [{gateway}] was not found.{bcolors.ENDC}')
            return

        try:
            rule_id = kwargs.get("rule_id")
            rules = Rules(record=gateway_context.configuration, params=params)
            rule_item = rules.get_rule_item(rule_type=RuleTypeEnum.ACTION, rule_id=rule_id)
            if rule_item is None:
                raise ValueError("Rule Id does not exist.")

            rule_action = kwargs.get("rule_action")
            if rule_action is not None:
                rule_item.action = RuleTypeEnum.find_enum(rule_action)
            priority = kwargs.get("priority")
            if priority is not None:
                rule_item.priority = priority
            ignore_case = kwargs.get("ignore_case")
            if ignore_case is not None:
                rule_item.case_sensitive = not ignore_case
            shared_folder_uid = kwargs.get("shared_folder_uid")
            if shared_folder_uid is not None:
                rule_item.shared_folder_uid = shared_folder_uid
            statement = kwargs.get("statement")
            if statement is not None:
                rule_item.statement = PAMGatewayActionDiscoverRuleAddCommand.validate_rule_statement(
                    params=params,
                    gateway_context=gateway_context,
                    statement=statement
                )
            rules.update_rule(rule_item)
            print(f"{bcolors.OKGREEN}Rule has been updated{bcolors.ENDC}")
        except Exception as err:
            print(f"{bcolors.FAIL}Rule was not updated: {err}{bcolors.ENDC}")



import io
from unittest import TestCase

from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey

from keepercommander import crypto, utils


class TestCrypto(TestCase):
    def test_stream_decrypter(self):
        data = bytearray(999)
        for i in range(len(data)):
            data[i] = i & 0xff

        key = utils.generate_aes_key()

        crypter = crypto.StreamCrypter()
        crypter.key = key
        crypter.is_gcm = False

        encrypted_data = bytearray()
        with crypter.set_stream(io.BytesIO(data), True) as cs:
            while True:
                buffer = cs.read(1024)
                if not buffer:
                    break
                encrypted_data.extend(buffer)
        self.assertEqual(crypter.bytes_read, len(data))

        decrypted_data = bytearray()
        with crypter.set_stream(io.BytesIO(encrypted_data), False) as cs:
            while True:
                buffer = cs.read(1024)
                if not buffer:
                    break
                decrypted_data.extend(buffer)

        self.assertEqual(decrypted_data, data)

    def test_decrypt_aes_v1(self):
        data = utils.base64_url_decode('KvsOJmE4JNK1HwKSpkBeR5R9YDms86uOb3wjNvc4LbUnZhKQtDxWifgA99tH2ZuP')
        key = utils.base64_url_decode('pAZmcxEoV2chXsFQ6bzn7Lop8yO4F8ERIuS7XpFtr7Y')
        data = crypto.decrypt_aes_v1(data, key)
        self.assertEqual(data, utils.base64_url_decode('6lf4FGVyhDRnRhJ91TrahjIW8lTqGA'))

    def test_encrypt_aes_v2(self):
        key = utils.base64_url_decode('c-EeCGlAO7F9QoJThlFBrhSCLYMe1H6GtKP-rezDnik')
        data = utils.base64_url_decode('nm-8mRG7xYwUG2duaOZzw-ttuqfetWjVIzoridJF0EJOGlDLs1ZWQ7F9mOJ0Hxuy' +
                                       'dFyojxdxVo1fGwbfwf0Jew07HhGGE5UZ_s57rQvhizDW3F3z9a7EqHRon0EilCbMhIzE')
        nonce = utils.base64_url_decode('Nt9_Y37C_43eRCRQ')
        enc_data = crypto.encrypt_aes_v2(data, key, nonce)
        expected_data = utils.base64_url_decode('Nt9_Y37C_43eRCRQCptb64zFaJVLcXF1udabOr_fyGXkpjpYeCAI7zVQD4JjewB' +
                                                'CP1Xp7D6dx-pxdRWkhDEnVhJ3fzezi8atmmzvf2ICfkDK0IHHB8iNSx_R1Ru8Tozb-IdavT3wKi7nKSJLDdt-dk-Mw7bCewpZtg4wY-1UQw')
        self.assertEqual(enc_data, expected_data)

        dec_data = crypto.decrypt_aes_v2(enc_data, key)
        self.assertEqual(dec_data, data)

    def test_encrypt_aes_v1(self):
        iv = utils.base64_url_decode('KvsOJmE4JNK1HwKSpkBeRw')
        block = utils.base64_url_decode('6lf4FGVyhDRnRhJ91TrahjIW8lTqGA')
        key = utils.base64_url_decode('pAZmcxEoV2chXsFQ6bzn7Lop8yO4F8ERIuS7XpFtr7Y')
        enc = crypto.encrypt_aes_v1(block, key, iv)
        encoded = utils.base64_url_encode(enc)
        self.assertEqual(encoded, 'KvsOJmE4JNK1HwKSpkBeR5R9YDms86uOb3wjNvc4LbUnZhKQtDxWifgA99tH2ZuP')

    def test_encrypt_rsa(self):
        data = crypto.get_random_bytes(100)
        puk = crypto.load_rsa_public_key(utils.base64_url_decode(_test_public_key))
        enc_data = crypto.encrypt_rsa(data, puk)
        prk = crypto.load_rsa_private_key(utils.base64_url_decode(_test_private_key))
        dec_data = crypto.decrypt_rsa(enc_data, prk)
        self.assertEqual(data, dec_data)

    def test_decrypt_rsa_short_ciphertext(self):
        def decrypt_message(encrypted_hex, apply_padding=False):
            data = bytes.fromhex(encrypted_hex)
            result = None
            try:
                plaintext_bytes = crypto.decrypt_rsa(data, rsa_key, apply_padding=apply_padding)
                result = plaintext_bytes.decode()
            finally:
                return result

        key_hex = ('308204a2020100028201007b3101c3c3d9e5c3e221c9bf2b12007f7c18a8b9732dd304760d0900221d0391a1c318410'
                   '74a52a9f5c73cda7114f042834990d7218d31601d838115b7dd5b6685502652f1453c940f499918781b335395658316'
                   '9795f9971c4431492378ca1e15f625e68b8a7678e4192b265e8517276397901517a6f0bc79a8367f5218346f3d8823a'
                   '336b1cd7370f76ecdc12f09fe92b8f0f723d086a64cdedc46af0c99ac5b5b050df9dd8b1cac21a529b3fd95086178a9'
                   'bcbfff3da92aa377aa344dca4469524594ef18061bb10452a309efdad9cc421ac5a9a833960708c32543c84d1064136'
                   '769c0e3156bc1b835ac17421a3dcb01b82b88ab59157f1bbe53c5a44a3702030100010282010068b2def0157d1464e5'
                   'c497a54ca2b11fa84580e8943666f88ca84974fe8930264e97f3fe1887173871b59247890225ac31ce8d35f9c2f92ad'
                   'a0a90e3f76f3f2623b959c8f65b44c0053a24ce820d8412ce8f06d9659dc611a2a96645e5cadbe4b3ff8e78a131ddbc'
                   'a307acffa02776e5382471052c23eac814915d37da7acddc5ebea2c9c84cabb8ad50ce297313131eacf52864bf2f546'
                   '078da3bed3bc03b7b3aaf881c17fff87b81ba201f21377dc73f67835da7179cafa273e7594a757d3ad7c6b6aed07b3a'
                   '993a0940514a7d8fe7b65492897dae020342126f0d95171fc33cb56205881df9a7b21d7b87c737902968b13d97a8593'
                   'b9973682b45595b34c902818100ecc10d7ebb81328272dd4dae731871e3649f5e5630260269ec91e5ccae9faec8459f'
                   '62c6bb4f950527e1c248b7a5ea5f3b8d1418d54735e1768bdb0b2b0472cee679a106d342cf7932c3495818e5484850e'
                   'cd2eb075743bb3c977b23aa1c9227fa99ab5612dc34824d78f6a86bc522a98115170beb8955e907aceff1594fe5b302'
                   '8181008534ab6d754cd5a7c6a94ac8f98b69c6118b7dabb8f5965291e16688613da8b6b60ae528897e82f02b0b7ee12'
                   'e89dec709333ae8438c0f69c3327ed8ba5243a00c5d55ec0b31d595d64ccaa51c0084db54d34ea7f13abe5deb06a1cd'
                   '02bb06d6c2bd1fd9a1478ae7f03c71b304a217ca012311bbdae1e586871923ee0e2e0f6d0281802b24f7379c25ec357'
                   '7873acbcaafaeb978b1ce3838a804929708f36ebc77df1b220cecac38a04510de76b6b817b785a17b31b772db13120f'
                   '9751df4606bdd5ca3c97f7af4dba84229b0c9986136b5d23c8938fd042d335459ec2202f9ca57e4108db0e2d2e5cb0b'
                   '8fa334c07df33daa03724c7c16557eeaefbb61937cb45d31f0281807c95f68039e5d32f48afae32aab3aa0a86fe605b'
                   'ec72465693faad5b81179a64c97f073612e330b4508e3fed7d099643b267280174abdafea082ea00eac3665c9b33f0d'
                   '904df6754ed4a857e47e274606fc5f31b409420d8d6a92d4c01f1cb43b28010fa0bce4e2d0094880357a2037dfbf240'
                   'f3e294c5883d735617a14b934102818100e8ebbea878b36ae483715b83255c6150443f1afff2cd105375df0b0448bab'
                   '92d957d455036b32fc25b36f926bb3f0aef2e60ead9d731911bc59109319fed35b11ffff3ca44656e64589774f25a4f'
                   '40c4c79adb954b74e0d0bf9b1923117fd70a755d28f56c1f6e5b148366d9136be4a07302d90d4f7fe8f0b1ccb1b617f'
                   '3ca83')
        key = bytes.fromhex(key_hex)
        rsa_key = crypto.load_rsa_private_key(key)  # type: RSAPrivateKey
        plaintext = 'This is a message'

        # The following value is a sample output of jsbn-rsa encryption, which can sometimes give us length < 2048-bit
        ciphertext_hex = ('ea6ef330b9e96b078374f892127e4adda7c27451abc3d4198e74b1a6a779afd82c9c1ca206bd6055f8238e3fa77b'
                          '62d7e9a6963a148f129c9371d9aad41a99d98b336e084f1b8e09e9f3f80595fc0991ce11269c57accb021307ec45'
                          '570a2061f7d7ddb3478eed0c57f464371eccfe90e13e78e7bc016cc5fd80624140497eb91d83a1d4661b8aa7c08d'
                          '7cc90373c891ccda46e2e01351c8944c2170e3aad46cfc97511469c10913169e33c3febccee5a2e0cfa5c8958741'
                          '3a3b7f712b89ffc60c8572d89b70ba45cc6158e96ffd37dc5f685863eff00df5ccdbe7c01ad91316286ee0793553'
                          '0015cbc13827ff1975b305cf132985b5f02fa7a8dd57d96533')

        # If no left-padding of 0s is applied to this ciphertext, we expect the decryption to fail
        self.assertIsNone(
            decrypt_message(ciphertext_hex, apply_padding=False)
        )
        # Otherwise, we expect it to give us the original plaintext message
        self.assertEqual(
            decrypt_message(ciphertext_hex, apply_padding=True),
            plaintext
        )

    def test_ec_encryption(self):
        private_key = crypto.load_ec_private_key(utils.base64_url_decode(_test_ec_private_key))
        decrypted_data = crypto.decrypt_ec(utils.base64_url_decode(_test_ec_encrypted_data), private_key)
        data = utils.base64_url_decode(_test_random_data)
        self.assertEqual(decrypted_data, data)

    def test_derive_key_hash_v1(self):
        password = 'q2rXmNBFeLwAEX55hVVTfg'
        salt = utils.base64_url_decode('Ozv5_XSBgw-XSrDosp8Y1A')
        iterations = 1000
        expected_key = utils.base64_url_decode('nu911pKhOIeX_lToXa4uIUuMPg1pj_3ZGpGmd7OjvRs')
        key_hash = crypto.derive_keyhash_v1(password, salt, iterations)
        self.assertEqual(key_hash, expected_key)

    def test_derive_key_hash_v2(self):
        password = 'q2rXmNBFeLwAEX55hVVTfg'
        salt = utils.base64_url_decode('Ozv5_XSBgw-XSrDosp8Y1A')
        iterations = 1000
        expected_key = utils.base64_url_decode('rXE9OHv_gcvUHdWuBIkyLsRDXT1oddQCzf6PrIECl2g')
        domain = '1oZZl0fKjU4'
        key_hash = crypto.derive_keyhash_v2(domain, password, salt, iterations)
        self.assertEqual(key_hash, expected_key)

    def test_password_score(self):
        self.assertEqual(utils.password_score('!@#$%^&*()'), 92)
        self.assertEqual(utils.password_score('aZkljfzsnmp4w9058dsqln5yf(&*))(*)(345'), 100)
        self.assertEqual(utils.password_score('c3>^sxuKZ[Ndyo(OBE14'), 100)
        self.assertEqual(utils.password_score('AAAbbbCCC11'), 38)
        self.assertEqual(utils.password_score('password'), 8)


_test_random_data = \
    'cKGoVph_X0NKjk8jQgxyQWRElUY7IsbbIJaRcJVlnOb7AchFiY-izmTTOlgArwIqAxKDKSRAWx2Q1pX' \
    'mWUUoEVKwFE9B2pLUi_GrTQu_hPEm6HjinwJaUPql-kqDfvGLiX6yBw'

_test_ec_private_key = 'HIIeyuuRkVGvhtax8mlX7fangaC6DKa2R8VAg5AAtBY'

_test_ec_encrypted_data = \
    'BMObx0eJO48N3Fg-2rQzVzppGi51kkxCM3KTRr2Zqe_q1kS8waisb2X9mDMnGKCtJqJLsVdTYsCn_BZ' \
    'zrbXO8846fMBqc1fKnOIBnxc_YIvTU9iKdx3vKgonOrPkyLMu273YQx8MpP-_FUjymRG7Rf4J0BIroP' \
    '5ndvaC6bzDmZGS-KLP8zO7oY7oNgeOI9Tj3AXCW65No2QzvGsppEHK1DcPpL8dh4RK2vt8Hh9Udj86Y' \
    'd-pi11AXusnzysl6cwzWA'

_test_public_key = "MIIBCgKCAQEAqR0AjmBXo371pYmvS1NM8nXlbAv5qUbPYuV6KVwKjN3T8WX5K6HD" \
                   "Gl3-ylAbI02vIzKue-gDbjo1wUGp2qhANc1VxllLSWnkJmwbuGUTEWp4ANjusoMh" \
                   "PvEwna1XPdlrSMdsKokjbP9xbguPdvXx5oBaqArrrGEg-36Vi7miA_g_UT4DKcry" \
                   "glD4Xx0H9t5Hav-frz2qcEsyh9FC0fNyon_uveEdP2ac-kax8vO5EeVfBzOdw-WP" \
                   "aBtUO1h7rSZ6xKOm6x1OahNTUFy7Cgm038JuMwHChTK29H9EOlqbOOuzYA1ENzL8" \
                   "8hELpe-kl4RmpNS94BJDssikFFbjoiAVfwIDAQAB"

_test_private_key = "MIIEogIBAAKCAQEAqR0AjmBXo371pYmvS1NM8nXlbAv5qUbPYuV6KVwKjN3T8WX5" \
                    "K6HDGl3-ylAbI02vIzKue-gDbjo1wUGp2qhANc1VxllLSWnkJmwbuGUTEWp4ANju" \
                    "soMhPvEwna1XPdlrSMdsKokjbP9xbguPdvXx5oBaqArrrGEg-36Vi7miA_g_UT4D" \
                    "KcryglD4Xx0H9t5Hav-frz2qcEsyh9FC0fNyon_uveEdP2ac-kax8vO5EeVfBzOd" \
                    "w-WPaBtUO1h7rSZ6xKOm6x1OahNTUFy7Cgm038JuMwHChTK29H9EOlqbOOuzYA1E" \
                    "NzL88hELpe-kl4RmpNS94BJDssikFFbjoiAVfwIDAQABAoIBABB9KW64ahMg7-ai" \
                    "FBtuFdSWNjZgvIkKxHHKGi0qMkUl4-JnpPHiJdnOTGeBhAPfMTJnYKfoKV14A4HC" \
                    "W0NcoFYenTxnvHV-A6bTZ6iFAmTyUp0SicOSEY3Hiov1OMppBpLkDuHe2TtpdK_c" \
                    "JLLerCVjYnN8DRqTpdmfsAkdonRseXyhRhwO6yFwVy9TEc9_OFuqGMOsy5_VIts6" \
                    "pG0saJJUQlOuLTxHwtPdloqjI8l3yMiDfXvJF2_epb_PYpKkAQZy_UWM5u4P_pnb" \
                    "UdImyYo6HBmnq-qO07J7b3yOSAzWhklBD7cMh1ucSOyF9-u03mLOfx2-SXq4tIuU" \
                    "Lz3RHZECgYEA0Rj-ipCKEPwQORViDFYYk1txzFSVKVX9Q-ozl6i93kTXx8GF7vkX" \
                    "L6SaEbKDA2EARuczr1gjymlvgRAwbsX7bDylSF6EsmPZ-EccNe4GoXmfbgMFDqGr" \
                    "3jVUmwEYwkte6EvP2Ha2GDwIuXFhcXWxgbbQxGGEcS5niei1mV0jv-sCgYEAzwv9" \
                    "BIYkeBC6_kejD2VwNzC1Jl97vg2It2URTZUGPFvcXh1Ed_i1itXwJ7wBjyBdwLJM" \
                    "IWjZcAYKET9NdBps2loATbOHrw4zFEqjKr_X-xSVU4bunipoY40fhl6a15ngUZ49" \
                    "3OJe_YtXEBHTVHorltIYuugu0zKk6uKbU_bt770CgYAR8_5u8UgZezr9W7umaYIE" \
                    "rPZRX_XKrcpoGWTCocdjnS-VxCT2xsZZ3d0opdYf5SU78T_7zyqLh4_-WeB-slsL" \
                    "CQ3777mfA3nEmn5ulvhUxveMX5AAmJsEIjoYcPiqPgRxF4lKAa9S11y8Z2LBdiR-" \
                    "ia7VHbZcbWqQab2l5FxcbwKBgCz_Ov7XtGdPo4QNx5daAVhNQqFTUQ5N3K-WzHri" \
                    "71cA09S0YaP9Ll88_ZN1HZWggB-X4EnGgrMA7QEwk8Gu2Idf1f8NDGj0Gg_H5Mwu" \
                    "o17S610azxMavlMcYYSPXPGMZJ74WBOAMwrBVKuOZDJQ1tZRVMSSH1MRB5xwoTdP" \
                    "TAi1AoGAXqJUfDAjtLR0wFoLlV0GWGOObKkPZFCbFdv0_CY2dk0nKnSsYRCogiFP" \
                    "t9XhZG5kawEtdfqiNBDyeNVLu6FaZnRkid_tUqMKfCYLjNDq31OD1Pwvyuh6Hs1P" \
                    "hL2-nt6t9b7JMyzKjWq_OPuTPH0QErL3oiFbTaZ4fDXplH_6Snw"
